; FIF81 - An Implementation of Forth 79 for the ZX81
; Copyright (C) 2021 Alan Bleasby

; This file is part of FIF81

; These files are free software: you can redistribute them and/or modify
; them under the terms of the GNU General Public License as published by
; the Free Software Foundation, either version 3 of the License, or
; (at your option) any later version.

; These files are distributed in the hope that they will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.

; You should have received a copy of the GNU General Public License
; along with this program.  If not, see <https://www.gnu.org/licenses/>.

; This version of FIF is for minimally upgraded ZX81s i.e. with 16K
; or more, possibly with a chr$128 UDG system and using cassette as the 
; storage medium.

; BC = Instruction register
; DE = Word Address register

#define xUDG			; Specify User defined graphics

#define	BREAK-1	 $0f46
#define KBD	 $02bb
#define DECODE	 $07bd
#define SET_FAST $02e7
#define	FAST	 $0f23
#define SLOW	 $0f2b
#define CLS	 $0a2a
#define	SAVE	 $02f6


#define LAST_K	$4025
#define	E_LINE	$4014
#define	S_POSN  $4039
#define	DF_CC	$400e
#define	D_FILE	$400c
#define FRAMES	$4034
#define	LBSIZE	$80		; Line buffer size
#define	STKSIZE $40		; Computation stack size
#define	WBSIZE	$40		; Word buffer size

#define CURU	$70
#define CURD	$71
#define	CURL	$72
#define	CURR	$73
#define	GRAPH	$74
#define	DELLINE	$75
#define	HALT	$76
#define NL	$0d	
#define	RUBOUT	$77
#define BS	$08

#define	CURSOR	$b1		; Alter if you want a different cursor.
				; Currently inverse L.
#define	KEYBIAS	$7e		; Displacement of key table from start of ROM
#define	DECODE_TAB Keytables-KEYBIAS
				; New Keytab location for DECODE
#define BLANK $20		; ASCII space

#define BUFSIZE $404		; Size of a screen in RAM
#define	BLKSIZE	$400

#define	CPL	$1c		; List: characters per line
#define LLINES	$24		; Number of (CPL) lines


	.org     $4009 
 
;= System variables ============================================ 

	.byte	0		; VERSN 
	.word	0     		; E_PPC 
	.word	Dfile		; D_FILE 
	.word	Dfile+1    	; DF_CC 
	.word	vars		; VARS 
	.word	0		; DEST 
	.word	last		; E_LINE 
	.word	last-1     	; CH_ADD 
	.word	0		; X_PTR 
	.word 	last		; STKBOT 
	.word	last		; STKEND 
	.byte 	0		; BERG 
	.word 	membot      	; MEM 
	.byte 	0     		; not used 
	.byte 	2		; DF_SZ 
	.word	1		; S_TOP 
	.byte	$FF,$FF,$FF     ; LAST_K 
	.byte	55		; MARGIN 
	.word	line10		; NXTLIN
	.word 	0		; OLDPPC 
	.byte	0		; FLAGX 
	.word	0		; STRLEN 
	.word 	$0C8D		; T_ADDR 
	.word 	0		; SEED 
	.word	$FFFF		; FRAMES 
	.byte	0,0		; COORDS 
	.byte	$BC		; PR_CC 
	.byte	33,24		; S_POSN 
	.byte 	01000000B  	; CDFLAG 
	.block	33		; Print buffer 

membot: 
	.block	30		; Calculators memory area 
	.block	2     		; not used 
 

line10
	.byte	0,10		; line number 
	.word	line20-$-2	; line length 
	.byte	$f9		; RAND
	.byte	$d4		; USR 
	.byte	$1c		; 0 
	.byte	$7e		; Float token 
	.byte	$8f		; Float number 
	.byte	$07
	.byte	$76
	.byte	$00
	.byte	$00
	.byte	HALT

line20
	.byte	$00,$20
	.word	Dfile-$-2
	.byte	$f7		; RUN
	.byte	HALT


Dfile
	.byte	HALT
	.word	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.byte	HALT
	.word	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.byte	HALT
	.word	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.byte	HALT
	.word	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.byte	HALT
	.word	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.byte	HALT
	.word	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.byte	HALT
	.word	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.byte	HALT
	.word	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.byte	HALT
	.word	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.byte	HALT
	.word	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.byte	HALT
	.word	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.byte	HALT
	.word	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.byte	HALT
	.word	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.byte	HALT
	.word	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.byte	HALT
	.word	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.byte	HALT
	.word	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.byte	HALT
	.word	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.byte	HALT
	.word	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.byte	HALT
	.word	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.byte	HALT
	.word	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.byte	HALT
	.word	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.byte	HALT
	.word	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.byte	HALT
	.word	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.byte	HALT
	.word	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.byte	HALT

vars	.byte	$80


	.block	$10		; Prevent vars crashing into the start

	jp	Start


DSP	.word	DStack		; Data stack pointer
Free	.block	12		; Allow 6 words of underflow
DStack	.block	STKSIZE		; 64 byte data stack

LBP	.word	LBuff		; Line buffer pointer
LBPsto	.word	$0000		; Line buffer ptr store for QUERY
LBPbase	.word	$00		; Start addr of current Line Buffer
LBuff	.block	LBSIZE		; 80 byte line buffer
	.byte	$80,$80		; Two line buff terminators
WBuff	.block	WBSIZE		; Word buffer
	.byte	$80,$80		; Word buffer terminators
	.block	$20		; Numeric part of PAD
PADL	.block	$60		; Text part of PAD

Flen	.byte	5		; Length of filename
Fname	.byte	$38,$28,$37	; 'SCR'.
	.byte	"xx",$80	; Forth screen base filename.
				; xx will be replaced by hex digits
AFname	.byte	5,"SCRxx",$80
Lname	.byte	$05,"SCR: "
Press	.byte	$0d,"HIT A KEY TO " 
SaveTxt .byte	$05,"SAVE "	
LoadTxt	.byte	$05,"LOAD "
PsvTxt	.byte	$2b,$2e,$2b+$80	; FIF filename for PSAVE
	

SysBlock
Context	.word	TOPNFA		; Offset 0
Current	.word	TOPNFA		; Offset 2
State	.byte	0		; Offset 4. False=exec True=compile
Mode	.byte	0		; Offset 5. 
Base	.byte	0		; Offset 6 Numeric base. 0 = cold start.
DP	.word	DicTop		; Offset 7 Dictionary pointer


UserBlock
Hld	.word	0			; Offset 0
Out	.word	0			; Offset 2
In	.word	0			; Offset 4
Blk	.word	0			; Offset 6
Tib	.word	0			; Offset 8
Scr	.word	0			; Offset A
Warn	.word	0			; Offset C
Rhash	.word	0			; Offset E


Tmp	.word	0
Ltmp	.byte	1
StkIni	.word	0
Scnt	.word	0		; Screen counter for ARROW
IsDub	.byte	0

Buff1	.block	BUFSIZE		; Screen buffers
Buff2	.block	BUFSIZE

BuffEnd


; Scan keyboard. Returns key pressed in 'A'
ScanKBD	.MODULE ScanKBD_M
	ld	a,(LAST_K)
	inc	a
	jr	nz,ScanKBD
_1	ld	bc,(LAST_K)
	ld	a,c
	inc	a
	jr	z,_1
	call	DECODE
	jr	nc,_1

	ld	de,DECODE_TAB
	add	hl,de
	ld	a,(hl)
	ret



; Read keyboard	and check ranges
ScanRepeat
	push	hl
	push	de
	push	bc
_srlp	call	ScanKBD

	cp	DELLINE
	jr	z,_srout
	cp	NL
	jr	z,_srout
	cp	BS
	jr	z,_srout

	cp	$20
	jr	c,_srlp
	cp	$a0
	jr	nc,_srlp

_srout	pop	bc
	pop	de
	pop	hl
	ret




RstMsg	.byte	$0b,"FIF RESTART"
BegMsg	.byte	$08,"FIF 1.03"
OKMSG	.byte	$03," OK"
QMSG	.byte	$02," ?"
StkMsg	.byte	$0f,"STACK UNDERFLOW"
NoMsg	.byte	$00




; Forth initialisation

Do_Forth	
	ld	hl,(StkIni)	; Reset return stack pointer
	ld	sp,hl

	call	EmpBufs		; Empty the screens buffers

	ld	de,RstMsg	; If BASE is not zero then set DE to be
	ld	a,(Base)	; a restart message
	and	a
	jr	nz,ABORT

	ld	a,$0a		; otherwise set numeric base to deimal
	ld	(Base),a	; and set DE to boot message
	ld	de,BegMsg
ABORT	ld	hl,DStack	; Reset computation stack
	ld	(DSP),hl
	ex	de,hl		; Push the address of the start message
	call	PushOne
	ld	hl,$0000
	ld	(State),hl	; Set to 'execute'
	ld	(Blk),hl	; Reset storage block number
	ld	hl,LBuff	; Reset TIB
	ld	(Tib),hl

	ld	hl,$8080	; Put end-of-buffer markers in line buffer
	ld	(LBuff+LBSIZE),hl
	ld	bc,OUTER	; Load instruction pointer with address of
	jp	NEXT		; outer interpreter and start FORTH




; Empty buffers to NULL. Add $8080 end-of-buffer markers
; and reset Prev & Use.
	
EmpBufs
	ld	hl,Buff1
	ld	(Use),hl
	ld	(Prev),hl
	ld	de,BUFSIZE*2
eblp	ld	(hl),0
	inc	hl
	dec	de
	ld	a,d
	or	e
	jr	nz,eblp

	ld	hl,$00ff
	ld	(Buff1),hl
	ld	(Buff2),hl
	
	ld	hl,Buff1
	ld	de,BUFSIZE-2
	add	hl,de
	ld	(hl),$80
	inc	hl
	ld	(hl),$80
	inc	hl
	add	hl,de
	ld	(hl),$80
	inc	hl
	ld	(hl),$80
	ret




; Standard FORTH inner interpreter

COLON	push	bc
	ld	b,d
	ld	c,e
	jr	NEXT
SEMI	.word	*+2
	pop	bc
NEXT	ld	a,(bc)
	ld	e,a
	inc	bc
	ld	a,(bc)
	ld	d,a
	inc	bc
RUN	ld	a,(de)
	ld	l,a
	inc	de
	ld	a,(de)
	ld	h,a
	inc	de
	jp	(hl)



; If the top stack entry is 0 then the instruction register is incremented by
; the value addressed by the instruction register, otherwise it is incremented
; by one to skip the displacemnt.

ZBRANCH	.word	*+2
	call	PopOne		; Get top stack item
	ld	a,h		; Is it zero?
	or	l
	jr	z,BRent		; Yes, so add jump byte
	inc	bc		; otherwise increment by one.
	jp	NEXT

	

	
; Increments the instruction register by the value whose address is held
; in the instruction register

	.MODULE BRANCH_M
BRANCH	.word	*+2
BRent	ld	h,b
	ld	l,c
	ld	e,(hl)
	ld	d,0
	bit	7,e
	jr	z,_1
	dec	d
_1	add	hl,de
	ld	b,h
	ld	c,l
	jp	NEXT




; Get start index and terminator from the data stack. Move both to the return 
; stack with the start index as the top entry

S_DO	.word	*+2
	call	PopOne
	ex	de,hl
	call	PopOne
	push	hl
	push	de
	jp	NEXT




; Increments the top return stack item and compares it with the second return
; stack entry. If the second is greater than the first a jump to BRANCH
; happens, otherwise the return stack items are dropped and IR incremented
; to skip the jump byte.

	.MODULE S_LOOP_M
S_LOOP	.word	*+2
	ld	e,$01		; Set increment to 1
D_LOOP	ld	d,$00
	bit	7,e
	jr	z,_1
	dec	d
_1	pop	hl		; Index
	and	a
	adc	hl,de
	ld	a,d
	pop	de		; limit
	jp	pe,_term
	push	de
	push	hl
	rlca
	jr	nc,_2
	ex	de,hl
_2	ld	a,h
	xor	d
	jp	m,_3
	sbc	hl,de
_3	rl	h
	ccf
	jp	nc,BRent
	pop	hl
	pop	hl
_term	inc	bc
	jp	NEXT




; Gets the DO increment value from the data stack and jumps to D_LOOP.
S_XLOOP	.word	*+2
	call	PopOne
	ld	e,l
	jr	D_LOOP




; Pushes to the stack the word whose address is in the instruction 
; register and increments the instruction register twice (past the literal)

LIT	.word	*+2
	ld	a,(bc)		; Get IR byte
	ld	l,a		; Move to HL
	inc	bc		; Bump IR
	ld	a,(bc)		; Get second byte
	ld	h,a		; Finish move to HL
	inc	bc		; Bump IR
	call	PushOne		; Push to stack
	jp	NEXT




; Pushes to the stack the byte whose address is in the instruction 
; register and increments the instruction register past the literal

CLIT	.word	*+2
	ld	a,(bc)		; Get byte at IR
	ld	l,a		; Move to HL
	inc	bc		; Bump IR
	rla			; Sign to Carry
	sbc	a,a		; $FF if neg else $00
	ld	h,a		; Set sign extension
	call	PushOne		; and push to stack
	jp	NEXT




; Outer interpreter defined by secondaries

OUTER
outec	.word	CTYPE
	.word	INLINE

	.word	INTER

	.word	QUESTION
	.word	BRANCH
outsc	.byte	outec-outsc




INTER	.word	COLON
intreb	.word	BL
	.word	TOKEN
	.word	QSEARCH
	.word	ZBRANCH
intrsa	.byte	intrea-intrsa
	.word	QEXECUTE
	.word	BRANCH
intrsc	.byte	intreb-intrsc
intrea	.word	QNUMBER
	.word	ZBRANCH
intrsb	.byte	intreb-intrsb
	.word	SEMI




QSEARCH
	.word	COLON
	.word	CONTEXT
	.word	AT
	.word	AT
	.word	SEARCH
	.word	ZEROEQ
qsea	.word	SEMI




QNUMBER
	.word	COLON
	.word	DOTCALC
	.word	ZBRANCH
qnmbad	.byte	qnm2ea-qnmbad

	.word	PARSENUM
	.word	ZBRANCH
qnm2sa	.byte	qnm2ea-qnm2sa
	.word	ROT
	.word	CAT
	.word	CLIT
	.byte	'.'
	.word	SUB

	.word	ZBRANCH
qnm2sb	.byte	qnm2eb-qnm2sb
	.word	DROP
	.word	LITER
	.word	ZERO
	.word	BRANCH
qnm2sc	.byte	qnm2ex-qnm2sc

qnm2eb	.word	DLITER
	.word	ZERO
	.word	BRANCH
qnm2sd	.byte	qnm2ex-qnm2sd
	

qnm2ea	.word	ONE
qnm2ex	.word	SEMI




QEXECUTE
	.word	COLON
	.word	MODE
	.word	CAT
	.word	ZEROEQ
	.word	ZBRANCH
qesa	.byte	qeea-qesa	

	.word	STATE
	.word	CAT
	.word	ZEROEQ
	.word	ZBRANCH
qesc	.byte	qeec-qesc	

qeea	.word	EXECUTE
	.word	STACK
	.word	BRANCH
qesb	.byte	qeeb-qesb
qeec	.word	COMMA
qeeb	.word	ZERO
	.word	MODE
	.word	CSTORE
	.word	SEMI




PARSENUM
	.word	COLON
	.word	DUP
	.word	CAT
	.word	OVER
	.word	PLUS
	.word	TOR
	.word	ZERO
	.word	ZERO
	.word	ROT
	.word	DUP
	.word	ONEP
	.word	CAT
	.word	CLIT
	.byte	'-'
	.word	EQUAL
	.word	DUP
	.word	TOR
	.word	PLUS
	.word	CONVERT
	.word	RFROM
	.word	ZBRANCH
n2sa	.byte	n2ea-n2sa
	.word	TOR
	.word	DNEGAT
	.word	RFROM
n2ea	.word	RFROM
	.word	OVER
	.word	SUB
	.word	DUP
	.word	ZLESS
	.word	ZBRANCH
n2sb	.byte	n2eb-n2sb
	.word	TDROP
	.word	ONE
	.word	BRANCH
n2sc	.byte	n2ex-n2sc
n2eb	.word	ZBRANCH
n2sd	.byte	n2ed-n2sd
	.word	TDROP
	.word	DROP
	.word	ZERO
	.word	BRANCH
n2se	.byte	n2ex-n2se
n2ed	.word	CAT
	.word	CLIT
	.byte	'.'
	.word	SUB
	.word	ZBRANCH
n2sf	.byte	n2ef-n2sf
	.word	TDROP
	.word	DROP
	.word	ZERO
	.word	BRANCH
n2sg	.byte	n2ex-n2sg
n2ef	.word	ONE
n2ex	.word	SEMI




; Pre-process numeric input of the form 123.456 to 123456. thereby
; allowing the dot to appear in any position except the first
; Multiple dots are disallowed. This enables easier numeric
; formatting by the user. This was a suggestion from Moggy.

	.MODULE	DOTCALC_M
DOTCALC	.word	*+2
	push	bc
	
	xor	a		; Set not double flag
	ld	(IsDub),a
	ld	hl,(DP)
	ld	a,(hl)
	ld	e,a		; DE = count
	ld	d,$00

	inc	hl
	ld	a,(hl)
	cp	'.'
	jr	z,_100

	add	hl,de
	ld	(hl),e		; New count pos	
	push	hl		; Save New digit start address
	inc	hl

	ld	b,e

	ld	de,(DP)
	inc	de		; DE=source, HL=Dest
	ex	de,hl		; HL=Source, DE=Dest

_1	ld	a,(hl)
	cp	'.'
	jr	nz,_2
	ld	a,(IsDub)
	inc	a
	ld	(IsDub),a
	jr	_3
_2	ld	(de),a
	inc	de
_3	inc	hl
	djnz	_1

	ex	de,hl

	ld	a,(IsDub)
	or	a
	jr	z,_4
	ld	(hl),'.'	; HL - new end
	inc	hl

_4	ld	(hl),0		; Make sure there are no more digits
	dec	hl

	call	PushOne		; Stack = End/Start/TRUE
	pop	hl
	call	PushOne
	ld	hl,$0001
	jr	_101


_100	ld	hl,$0000	; Failure
_101	call	PushOne

	pop	bc
	jp	NEXT





; Get top of data stack into HL
PopOne
	push	de
	ld	hl,(DSP)
	dec	hl
	ld	e,(hl)
	dec	hl
	ld	d,(hl)
	ld	(DSP),hl
	ex	de,hl
	pop	de
	ret




; Push HL to top of data stack
PushOne
	push	hl
	push	de
	ex	de,hl
	ld	hl,(DSP)
	ld	(hl),d
	inc	hl
	ld	(hl),e
	inc	hl
	ld	(DSP),hl
	pop	de
	pop	hl
	ret




; Nfa2Lfa
; Enters with nfa in HL, exits with lfa  in HL

Nfa2Lfa
	ld	a,(hl)		; Go to link field as above
	and	$1f		; Mask off Exec/Imm bits
	ld	d,$00
	ld	e,a
	scf
	adc	hl,de
	ret




; Get input for line buffer


	.MODULE INLINE_M
INLINE	.word	*+2
	push	bc
	call	LBInput
	pop	bc
	jp	NEXT




	.MODULE QUESTION_M

QUESTION
	.word	*+2
	ld	hl,(DP)		; Get ptr to dictionary
	inc	hl		; Step over length
	xor	a		; Test for NULL terminator
	cp	(hl)
	jr	z,_isterm
	bit	7,(hl)		; If bit set, a $80 terminator
	jr	z,_error	; If not set, an error
_isterm	ld	hl,$0000	; Reset TIB offset
	ld	(In),hl
	ld	hl,OKMSG	; Put OK msg on stack
	call	PushOne
	jp	NEXT
_error	call	CRLF		; Not a number of dictionary word
	dec	hl		; so print offending word and '?'
	ld	e,(hl)
_1	inc	hl
	ld	a,(hl)
	call	PrtChr
	dec	e
	jr	nz,_1
	ld	de,QMSG
	jr	PATCH




; Reset some sysvars if there's been an error.
; Delete any partial word definition.

PATCH	ld	a,(State)	; Get state variable
	and	a		; Is it zero (execute)?
	jp	z,ABORT		; If so, goto restart
	push	de		; else save message address
	ld	hl,(Current)	; Get vocabulary address
	ld	e,(hl)		; It points to the latest
	inc	hl		; entry which was aborted
	ld	d,(hl)		; This is where DP
	ex	de,hl		; should point
	ld	(DP),hl		; Restore DP
	push	de
	call	Nfa2Lfa		; Point to start of LFA
	pop	de
	inc	hl		; HL = second byte of link field

	ld	a,(hl)		; Move link address to the
	ld	(de),a		; current vocabulary as the pointer
	dec	hl		; to its latest entry
	dec	de
	ld	a,(hl)
	ld	(de),a
	pop	de		; Restore msg address and exit to
	jp	ABORT		; restart.




; Check for stack underflow
STACK	.word	*+2
	ld	hl,(DSP)	; Subtract data stack start from DSP
	ld	de,DStack	; If C set then an underflow
	and	a		; has happened.
	sbc	hl,de
	jp	nc,NEXT		; No overflow
	ld	hl,DStack	; Overflow, so reset DSP and jump to PATCH
	ld	(DSP),hl
	ld	de,StkMsg
	jp	PATCH




; ABORT

ABORTH	.byte	$85,"ABORT"
	.word	$00
ABRT	.word	*+2
	jp	Do_Forth




; QUIT

QUITH	.byte	$84,"QUIT"
	.word	ABORTH
QUIT	.word	*+2
	xor	a
	ld	(State),a	; Reset execute mode
	ld	hl,$000		; Reset block to Line Buff
	ld	(Blk),hl
	ld	hl,(StkIni)	; Reset return stack
	ld	sp,hl
	ld	hl,NoMsg	; Dummy message
	call	PushOne
	ld	bc,OUTER	; Back to outer interpreter
	jp	NEXT




; Execute user machine code

CALLH	.byte	$84,"CALL"
	.word	QUITH
CALLM	.word	*+2
	call	PopOne
	ld	(_callc+1),hl
_callc	call	$0000
	jp	NEXT




; Execute routine at top of data stack

EXECUTEH
	.byte	$87,"EXECUTE"
	.word	CALLH
EXECUTE	.word	*+2
	call	PopOne
	ex	de,hl
	jp	RUN




;  Search dictionary for keyword at DP
; If a match is found then the CFA of the matching word is stacked followed
; by a FALSE flag. If no match is found then a TRUE flag is stacked.
; The flag values are counter-intuitive but make the code shorter later.
SEARCHH	.MODULE SEARCH_M
	.byte	$86,"SEARCH"
	.word	EXECUTEH
SEARCH	.word	*+2	
	call	S_SEARCH
	jp	NEXT



; SEARCH is just a header that calls this routine to do all the work
S_SEARCH
	push	bc
	push	de

	ld	hl,Mode		; Set mode to indicate word not immediate
	ld	(hl),0

	call	PopOne		; Get NFA at end of list from data stack

_sloop	push	hl		; Save start of header
	ld	de,(DP)		; Get keyword end-of-dictionary address
	ld	a,(hl)		; Get length byte of keyword
	bit	5,a		; Smudge bit
	jr	nz,_nxtnfa	; Smudged
	ld	a,(de)
	and	$1f		; Mask off high order bits
	ld	b,a		; Length to B
	ld	a,(hl)		; Get dictionary length byte
	and	$1f		; and mask off high order bits
	cp	b		; Compare lengths
	jr	nz,_nxtnfa	; Length mismatch

_nxtch	inc	hl		; Loop round comparing all characters
	inc	de
	ld	a,(de)	
	ld	c,(hl)		; Dictionary char to C
	cp	c		; Compare characters
	jr	nz,_nxtnfa	; Mismatch found
	djnz	_nxtch		; Loop for next char

	pop	hl		; Start of found header

	bit	6,(hl)		; Test whether Immediate and set flag if so
	jr	z,_1		; Not immediate

	ld	a,$01		; Immediate so set 'Mode'
	ld	(Mode),a

_1	call	Nfa2Lfa
	inc	hl		; Go to cfa
	inc	hl
	call	PushOne		; Put cfa on data stack
	ld	c,$00		; Clear flag (B already zero)
	jr	_flag

_nxtnfa	pop	hl
	call	Nfa2Lfa
	ld	e,(hl)		; Get link address
	inc	hl
	ld	d,(hl)
	ex	de,hl
	ld	a,h		; Compare link address to zero
	or	l
	jr	nz,_sloop	; Not at end of list
	ld	bc,$01		; End so push a true flag

_flag	ld	h,b		; Push flag to data stack.
	ld	l,c
	call	PushOne
	pop	de
	pop	bc
	ret


; Copy next word in line buffer to the top of the dictionary (preceded by a
; length byte) and search the dictionary for a match (starting from the
; stacked NFA.
S_FIND	.word	*+2
	ld	hl,BLANK
	call	PushOne
	call	TokenMC
	call	S_SEARCH
	jp	NEXT




; FIND
; Stack the NFA of the top word in the context vocabulary,
; copy the next word in line buffer to the top of the dictionary (preceded
; by a length byte) and search the dictionary for a match.

FINDH	.byte	$84,"FIND"
	.word	SEARCHH
FIND	.word	COLON
	.word	CONTEXT
	.word	AT
	.word	AT
	.word	S_FIND
	.word	ZBRANCH
finsa	.byte	finea-finsa
	.word	ZERO
finea	.word	SEMI




; TICK
#IFDEF UDG
TICKH	.byte	$81,$27
#ELSE
TICKH	.byte	$84,"TICK"
#ENDIF
	.word	FINDH
TICK	.word	COLON
	.word	FIND
	.word	DUP
	.word	ZEROEQ
	.word	ZBRANCH
ticksa	.byte	tickea-ticksa
	.word	QUESTION
tickea	.word	TWOP
	.word	LITER
	.word	SEMI




; TOKEN
; Move next token in line buffer to the DP address
; preceded by a length byte.
	.MODULE TOKEN_M
TOKENH	.byte	$85,"TOKEN"
	.word	TICKH
TOKEN	.word	*+2	
	call	TokenMC
	jp	NEXT




; Token code body (see above). Shared with S_FIND and WORD.
TokenMC .MODULE TokenMC_M
	push	bc
	push	de

	call	TokenParse	; Token start address is now in HL
	ld	de,(DP)		; Dict ptr
	ld	a,b		; Move count to A
	ld	(de),a		; Length to dictionary
	inc	de		; Bump dictionary address
	ld	c,b		; Get count to BC
	ld	b,0
	ldir			; Move token to dictionary
	pop	de
	pop	bc
	ret




; Parse token from LBP. Code shared with (WORD)

TokenParse
	call	PopOne		; Separator in C
	ld	b,h		; B is zero
	ld	c,l

	ld	hl,(LBP)	; Token ptr

	ld	a,c		; Skip separator bytes
_1	cp	(hl)
	jr	nz,_2
	inc	hl
	jr	_1

_2	push	hl		; Save start address
_3	or	a		; NULL terminator?
	jr	z,_4
	rla			; $80 terminator?
	jr	c,_4
	inc	b
	inc	hl		; Bump ptr
	ld	a,(hl)
	cp	c
	jr	nz,_3
	inc	hl		; Skip over separator
_4	ld	(LBP),hl


	ld	de,LBuff	; Change according to Blk
	ld	a,(Blk)		; If 0 then TIB else PREV
	or	a
	jr	z,_5
	ld	de,(Prev)

_5	and	a
	sbc	hl,de
	ld	(In),hl

	pop	hl		; Start address
	ret




; EXIT

EXITH	.byte	$84,"EXIT"
	.word	TOKENH
EXIT	.word	*+2
	pop	bc
	jp	NEXT




; DEPTH

DEPTHH	.byte	$85,"DEPTH"
	.word	EXITH
DEPTH	.word	*+2
	ld	hl,(DSP)
	ld	de,DStack
	and	a
	sbc	hl,de
	srl	h
	rr	l
	call	PushOne
	jp	NEXT




; FILL
	.MODULE FILL_M
FILLH	.byte	$84,"FILL"
	.word	DEPTHH
FILL	.word	*+2
	push	bc
	call	PopOne
	ld	c,l
	call	PopOne
	ex	de,hl
	call	PopOne
	bit	7,d
	jr	nz,_2
	ld	a,d
	or	e
	jr	z,_2

_1	ld	(hl),c
	inc	hl
	dec	de
	ld	a,d
	or	e
	jr	nz,_1
_2	pop	bc
	jp	NEXT




; MOVE

MOVEH	.byte	$84,"MOVE"
	.word	FILLH
MOVE	.word	COLON
	.word	DUP
	.word	PLUS
	.word	CMOVE
	.word	SEMI




; NOT

NOTH	.byte	$83,"NOT"
	.word	MOVEH
NOT	.word	COLON
	.word	ZEROEQ
	.word	SEMI




; -TRAILING

DTRAILH	.byte	$89,"-TRAILING"
	.word	NOTH
DTRAIL	.word	COLON
	.word	DUP
	.word	ZLESS
	.word	ZBRANCH
dtrasa	.byte	dtraea-dtrasa
	.word	QUESTION
dtraea	.word	DUP
	.word	ZERO
	.word	S_DO
dtralps	.word	TDUP
	.word	PLUS
	.word	ONEM
	.word	CAT
	.word	BL
	.word	SUB
	.word	ZBRANCH
dtrasb	.byte	dtraeb-dtrasb
	.word	LEAVE
	.word	BRANCH
dtrasc	.byte	dtraec-dtrasc
dtraeb	.word	ONEM
dtraec	.word	S_LOOP
dtralpe	.byte	dtralps-dtralpe
	.word	SEMI




; FORGET

FORGETH	.byte	$86,"FORGET"
.word	DTRAILH
FORGET	.word	COLON
	.word	CURRENT
	.word	AT
	.word	CONTEXT
	.word	STORE
	.word	TICK
	.word	DUP
	.word	LFA
	.word	AT
	.word	CURRENT
	.word	AT
	.word	STORE
	.word	NFA
	.word	DPTR
	.word	STORE
	.word	SEMI




; VLIST
	.MODULE VLIST_M
VLISTH	.byte	$85,"VLIST"
	.word	FORGETH
VLIST	.word	*+2
	push	bc
	ld	hl,(Context)
	ld	e,(hl)
	inc	hl
	ld	d,(hl)
	ex	de,hl

_vllp	ld	a,(hl)
	and	$3f		
	ld	b,a
_1	inc	hl
	ld	a,(hl)
	call	PrtChr
	djnz	_1	

	call	BREAK-1
	jr	c,_3
_2	call	BREAK-1
	jr	nc,_2

_3	ld	a,BLANK
	call	PrtChr
	ld	a,BLANK
	call	PrtChr


	inc	hl
	ld	e,(hl)
	inc	hl
	ld	d,(hl)

	ld	a,d
	or 	e
	jr	z,_out
	ex	de,hl
	jr	_vllp

_out	pop	bc
	jp	NEXT




; VOCABULARY

VOCABH	.byte	$8a,"VOCABULARY"
	.word	VLISTH
VOCAB	.word	COLON
	.word	CREATE
	.word	LAST
	.word	COMMA
	.word	SCODE
vocent	.byte	$cd,DoDoes%256,DoDoes/256
	.word	CONTEXT
	.word	STORE
	.word	SEMI




; FORTH

FORTHH	.byte	$85,"FORTH"
	.word	VOCABH
FORTH	.word	vocent
TOPNFA	.word	TopNfa




; Pops the top stack entry which points to a string consisting of a length
; byte followed by the string characters. Outputs the string.

CTYPEH	.byte	$85,"CTYPE"
	.word	FORTHH
CTYPE	.word	*+2
	call	PopOne		; Get string address
D_CTYPE	call	CTPrt
	jp	NEXT




CTPrt	.MODULE CTPrt_M
	ld	e,(hl)		; Get length
	xor	a
	cp	e
	ret	z
_1	inc	hl		; Bump ptr
	ld	a,(hl)		; Get character
	call	PrtChr		; Output character
	dec	e		; Decrement length
	jr	nz,_1		; Loop until length = 0.
	ret




; Code field routine for variables. Push the word at W to the stack
DoCon	ex	de,hl
	ld	e,(hl)
	inc	hl
	ld	d,(hl)
	ex	de,hl
	call	PushOne
	jp	NEXT




; Code field routine for variables. Push the W address to the stack
DoVar	ex	de,hl
	call	PushOne
	jp	NEXT
	



; Variable holding the CONTEXT address

CONTEXTH
	.byte	$87,"CONTEXT"
	.word	CTYPEH
CONTEXT	.word	DoSys
	.byte	0




; Variable holding the CURRENT address
CURRENTH
	.byte	$87,"CURRENT"
	.word	CONTEXTH
CURRENT	.word	DoSys
	.byte	2




; LAST

LASTH	.byte	$84,"LAST"
	.word	CURRENTH
LAST	.word	COLON
	.word	CURRENT
	.word	AT
	.word	AT
	.word	SEMI




; NEXT

NXTH	.byte	$84,"NEXT"
	.word	LASTH
NXT	.word	*+2
	ld	hl,(DP)
	ld	(hl),$c3
	inc	hl
	ld	(hl),NEXT%256
	inc	hl
	ld	(hl),NEXT/256
	inc	hl
	ld	(DP),hl
	jp	NEXT




; (;CODE)

SCODEH	.byte	$87,'(',$3b,"CODE)"
	.word	NXTH
SCODE	.word	COLON
	.word	RFROM
	.word	LAST
	.word	PFA
	.word	CFA
	.word	STORE
	.word	SEMI




; ;CODE


SCCODEH	.byte	$c5,$3b,"CODE"
	.word	SCODEH
SCCODE	.word	COLON
	.word	COMP
	.word	SCODE
	.word	LBRACE
	.word	SEMI




; VARIABLE

VARH	.byte	$88,"VARIABLE"
	.word	SCCODEH
VAR	.word	COLON
	.word	CREATE
	.word	ZERO
	.word	COMMA
	.word	SCODE
	jr	conste




; CONSTANT

CONSTH	.byte	$88,"CONSTANT"
	.word	VARH
CONST	.word	COLON
	.word	CREATE
	.word	COMMA
	.word	SCODE
	ex	de,hl
	ld	e,(hl)
	inc	hl
	ld	d,(hl)
conste	ex	de,hl
	call	PushOne
	jp	NEXT




; CMOVE
	.MODULE	CMOVE_M
CMOVEH	.byte	$85,"CMOVE"
	.word	CONSTH
CMOVE	.word	*+2
	push	bc
	call	PopOne
	ld	b,h
	ld	c,l
	call	PopOne
	ex	de,hl
	call	PopOne

	ld	a,b
	or	c
	jr	z,_1
	bit	7,b
	jr	nz,_1
	ldir
	pop	bc
_1	jp	NEXT




; (  - A  comment, terminated by ')' or term
	.MODULE BRACK_M
#IFDEF UDG
BRACKH	.byte	$c1,"("
#ELSE
BRACKH	.byte	$c2,"(("
#ENDIF
	.word	CMOVEH
BRACK	.word	*+2
#IFDEF UDG
	ld	hl,(LBP)
_1	ld	a,(hl)
	bit	7,a
	jr	nz,_2
	cp	')'
	jr	z,_2
	inc	hl
	jr	_1

_2	inc	hl
	ld	(LBP),hl
	jp	NEXT
#ELSE
	ld	hl,(LBP)
_1	ld	a,(hl)
_2	bit	7,a
	jr	nz,_4
	cp	')'
	jr	z,_3
	inc	hl
	jr	_1

_3	inc	hl
	ld	a,(hl)
	cp	')'
	jr	nz,_2
	inc	hl

_4	ld	(LBP),hl
	jp	NEXT
#ENDIF




; @ (? on non-UDG ZX81)
#IFDEF UDG
ATH	.byte	$81,"@"
#ELSE
ATH	.byte	$81,"?"
#ENDIF
	.word	BRACKH
AT	.word	*+2
	call	PopOne
	ld	e,(hl)
	inc	hl
	ld	d,(hl)
	ex	de,hl
PushNxt	call	PushOne
	jp	NEXT




; DUP

DUPH	.byte	$83,"DUP"
	.word	ATH
DUP	.word	*+2
	call	PopOne
	call	PushOne
	jr	PushNxt




; 2DUP

TDUPH	.byte	$84,"2DUP"
	.word	DUPH
TDUP	.word	COLON
	.word	OVER
	.word	OVER
	.word	SEMI




; ?DUP

QDUPH	.byte	$84,"?DUP"
	.word	TDUPH
QDUP	.word	*+2
	call	PopOne
	call	PushOne
	ld	a,h
	or	l
	jr	nz,PushNxt
	jp	NEXT




; Routine for System variables. Adds offset to Sys block address
; and pushes to stack
	
	.MODULE DoSys_M
DoSys	ld	hl,SysBlock
SDoSys	ld	a,(de)		; @W = offset
	add	a,l
	ld	l,a
	jr	nc,_1
	inc	h
_1	call	PushOne
	jp	NEXT




	.Module	DoUser_M
DoUser
	ld	hl,UserBlock
	jr	SDoSys




; STATE

STATEH	.byte	$85,"STATE"
	.word	QDUPH
STATE	.word	DoSys
	.byte	$04




; MODE

MODEH	.byte	$84,"MODE"
	.word	STATEH
MODE	.word	DoSys
	.byte	$05




; BASE

BASEH	.byte	$84,"BASE"
	.word	MODEH
BASE	.word	DoSys
	.byte	$06




; DP

DPH	.byte	$82,"DP"
	.word	BASEH
DPTR	.word	DoSys
	.byte	$07




; C@ (C? on the non-UDG ZX81)
#IFDEF UDG
CATH	.byte	$82,"C@"
#ELSE
CATH	.byte	$82,"C?"
#ENDIF
	.word	DPH
CAT	.word	*+2
	call	PopOne
	ld	e,(hl)
	ld	d,$00
	ex	de,hl
	call	PushOne
	jp	NEXT




; ,

COMMAH	.byte	$81,","
	.word	CATH
COMMA	.word	*+2
	call	PopOne		; Get word
	ex	de,hl
	ld	hl,(DP)		; Get DP
	ld	(hl),e		; Store LSB
	inc	hl		; Bump DP
	ld	(hl),d		; Store MSB
	inc	hl		; Bump DP and
	ld	(DP),hl		; update it
	jp	NEXT




; C,

CCOMMAH	.byte	$82,"C,"
	.word	COMMAH
CCOMMA	.word	*+2
	call	PopOne		; Get byte
	ex	de,hl
	ld	hl,(DP)		; Get DP
	ld	(hl),e		; Store byte
	inc	hl		; Bump DP and
	ld	(DP),hl		; update it
	jp	NEXT




; 2-

TWOMH	.byte	$82,"2-"
	.word	CCOMMAH
TWOM	.word	*+2
	call	PopOne
	dec	hl
	jr	TwoMen




; 1-

ONEMH	.byte	$82,"1-"
	.word	TWOMH
ONEM	.word	*+2
	call	PopOne
TwoMen	dec	hl
	call	PushOne
	jp	NEXT




; 1+

ONEPH	.byte	$82,"1+"
	.word	ONEMH
ONEP	.word	*+2
	call	PopOne
TwoPen	inc	hl
	call	PushOne
	jp	NEXT




; 2+

TWOPH	.byte	$82,"2+"
	.word	ONEPH
TWOP	.word	*+2
	call	PopOne
	inc	hl
	jr	TwoPen




; 0

ZEROH	.byte	$81,"0"
	.word	TWOPH
ZERO	.word	DoCon
	.word	$00




; 1

ONEH	.byte	$81,"1"
	.word	ZEROH
ONE	.word	DoCon
	.word	1




; 2

TWOH	.byte	$81,"2"
	.word	ONEH
TWO	.word	DoCon
	.word	2




; -1

MINUS1H	.byte	$82,"-1"
	.word	TWOH
TRUE
MINUS1	.word	DoCon
	.word	-1



; -2

MINUS2H	.byte	$82,"-2"
	.word	MINUS1H
MINUS2	.word	DoCon
	.word	-2




; =
	.MODULE EQUAL_M
EQUALH	.byte	$81,"="
	.word	MINUS2H
EQUAL	.word	*+2
	call	PopOne
	ex	de,hl
	call	PopOne
	and	a
	sbc	hl,de
	ld	hl,$0000
	jr	nz,_out
	inc	l
_out	call	PushOne
	jp	NEXT




; XOR

XORH	.byte	$83,"XOR"
	.word	EQUALH
XOR	.word	*+2
	call	PopOne
	ex	de,hl
	call	PopOne
	ld	a,e
	xor	l
	ld	l,a
	ld	a,d
	xor	h
	ld	h,a
	call	PushOne
	jp	NEXT



; OR

ORH	.byte	$82,"OR"
	.word	XORH
OR	.word	*+2
	call	PopOne
	ex	de,hl
	call	PopOne
	ld	a,e
	or	l
	ld	l,a
	ld	a,d
	or	h
	ld	h,a
	call	PushOne
	jp	NEXT




; AND

ANDH	.byte	$83,"AND"
	.word	ORH
AND	.word	*+2
	call	PopOne
	ex	de,hl
	call	PopOne
	ld	a,e
	and	l
	ld	l,a
	ld	a,d
	and	h
	ld	h,a
	call	PushOne
	jp	NEXT




; 0<
	.MODULE ZLESS_M
ZLESSH	.byte	$82,"0<"
	.word	ANDH
ZLESS	.word	*+2
	call	PopOne
	ld	de,$0000
	bit	7,h
	jr	z,_1
	inc	de
_1	ex	de,hl
	call	PushOne
	jp	NEXT




; 0>

ZGTRH	.byte	$82,"0>"
	.word	ZLESSH
ZGTR	.word	COLON
	.word	NEGATE
	.word	ZLESS
	.word	SEMI




; 0=
	.MODULE ZEROEQ_M
ZEROEQH	.byte	$82,"0="
	.word	ZGTRH
ZEROEQ	.word	*+2
	call	ZeroEQ
	jp	NEXT

ZeroEQ
	call	PopOne
	ld	de,$00
	ld	a,h
	or	l
	jr	nz,_1
	inc	de
_1	ex	de,hl
	call	PushOne
	ret

; <
	.MODULE	LESS_M
LESSH	.byte	$81,"<"
	.word	ZEROEQH
LESS	.word	*+2
	call	PopOne
	ex	de,hl
	call	PopOne
ltent	and	a
	sbc	hl,de
	ld	de,$0000
	jp	p,_1
	inc	e
_1	ex	de,hl
	call	PushOne
	jp	NEXT




; >
	.MODULE	GREAT_M
GREATH	.byte	$81,">"
	.word	LESSH
GREAT	.word	*+2
	call	PopOne
	ex	de,hl
	call	PopOne
	ex	de,hl
	jr	ltent




; MIN

MINH	.byte	$83,"MIN"
	.word	GREATH
MIN	.word	COLON
	.word	TDUP
	.word	GREAT
	.word	ZBRANCH
mnsa	.byte	mnse-mnsa
	.word	SWAP
mnse	.word	DROP
	.word	SEMI




; MAX

MAXH	.byte	$83,"MAX"
	.word	MINH
MAX	.word	COLON
	.word	TDUP
	.word	LESS
	.word	ZBRANCH
mxsa	.byte	mxse-mxsa
	.word	SWAP
mxse	.word	DROP
	.word	SEMI




; +-
PMH	.byte	$82,"+-"
	.word	MAXH
PM	.word	COLON
	.word	ZLESS
	.word	ZBRANCH
pmsa	.byte	pmea-pmsa
	.word	NEGATE
pmea	.word	SEMI




; D+-

DPMH	.byte	$83,"D+-"
	.word	PMH
DPM	.word	COLON
	.word	ZLESS
	.word	ZBRANCH
dpmsa	.byte	dpmea-dpmsa
	.word	DNEGAT
dpmea	.word	SEMI




; ABS

ABSH	.byte	$83,"ABS"
	.word	DPMH
ABS	.word	COLON
	.word	DUP
	.word	PM
	.word	SEMI



; DABS

DABSH	.byte	$84,"DABS"
	.word	ABSH
DABS	.word	COLON
	.word	DUP
	.word	DPM
	.word	SEMI




; NEGATE
NEGATEH	.byte	$06,"NEGATE"
	.word	DABSH
NEGATE	.word	*+2
	call	PopOne
	and	a
negatj	ld	de,$0000
	ex	de,hl
	sbc	hl,de
	call	PushOne
	jp	NEXT




; DNEGATE
DNEGATH	.byte	$87,"DNEGATE"
	.word	NEGATEH
DNEGAT	.word	*+2
	call	PopOne
	push	hl
	call	PopOne
	ld	de,$0000
	ex	de,hl
	and	a
	sbc	hl,de
	push	af
	call	PushOne
	pop	af
	ld	hl,$0000
	pop	de
	sbc	hl,de
	call	PushOne
	jp	NEXT





; DROP

DROPH	.byte	$84,"DROP"
	.word	DNEGATH
DROP	.word	*+2
	ld	hl,(DSP)
drope	dec	hl
	dec	hl
	ld	(DSP),hl
	jp	NEXT




; 2DROP

TDROPH	.byte	$85,"2DROP"
	.word	DROPH
TDROP	.word	*+2
	ld	hl,(DSP)
	dec	hl
	dec	hl
	jr	drope




; OVER
OVERH	.byte	$84,"OVER"
	.word	TDROPH
OVER	.word	*+2
	ld	hl,(DSP)
	push	hl
	dec	hl
	dec	hl
	dec	hl
	ld	e,(hl)
	dec	hl
	ld	d,(hl)
	pop	hl
	ld	(hl),d
	inc	hl
	ld	(HL),e
	inc	hl
	ld	(DSP),hl
	jp	NEXT




; PICK

PICKH	.byte	$84,"PICK"
	.word	OVERH
PICK	.word	*+2
	call	PopOne
	xor	a
	or	l
	jp	z,NEXT

	push	bc
	ld	b,l
	ld	hl,(DSP)

pcklp	dec	hl
	dec	hl
	djnz	pcklp
	ld	d,(hl)
	inc	hl
	ld	e,(hl)
	ld	hl,(DSP)
	ld	(hl),d
	inc	hl
	ld	(hl),e
	inc	hl
	ld	(DSP),hl
	pop	bc
	jp	NEXT




; ROLL
ROLLH	.byte	$84,"ROLL"
	.word	PICKH
ROLL	.word	*+2
	call	PopOne
	ld	a,$01
	cp	l
	jp	nc,NEXT

	push	bc

	ex	de,hl		; Length to DE
	sla	e		; *2
	rl	d

	ld	hl,(DSP)	; Add to stack pointer
	and	a
	sbc	hl,de

	ld	b,(hl)
	inc	hl
	ld	c,(hl)
	push	bc
	ld	b,0
	ld	c,e
	dec	c
	dec	c
	dec	hl
	ld	d,h
	ld	e,l
	inc	hl
	inc	hl
	ldir
	ld	hl,(DSP)
	dec	hl
	pop	bc
	ld	(hl),c
	dec	hl
	ld	(hl),b
	pop	bc
	jp	NEXT




; ROT
ROTH	.byte	$83,"ROT"
	.word	ROLLH
ROT	.word	*+2
	push	bc
	call	PopOne
	ex	de,hl		; DE=n3
	call	PopOne
	ld	b,h		; BC=n2
	ld	c,l
	call	PopOne		; HL=n1
	push	hl
	ld	h,b
	ld	l,c
	call	PushOne
	ex	de,hl
	call	PushOne
	pop	hl
	call	PushOne
	pop	bc
	jp	NEXT




; SWAP
SWAPH	.byte	$84,"SWAP"
	.word	ROTH
SWAP	.word	*+2
	call	PopOne
	ex	de,hl
	call	PopOne
	ex	de,hl
	call	PushOne
	ex	de,hl
	call	PushOne
	jp	NEXT




; >R
TORH	.byte	$82,">R"
	.word	SWAPH
TOR	.word	*+2
	call	PopOne
	push	hl
	jp	NEXT




; R>

RFROMH	.byte	$82,"R>"
	.word	TORH
RFROM	.word	*+2
	pop	hl
rfrme	call	PushOne
	jp	NEXT




; R@
#IFDEF UDG
RATH	.byte	$82,"R@"
#ELSE
RATH	.byte	$82,"R?"
#ENDIF

	.word	RFROMH
RAT	.word	*+2
	pop	hl
	push	hl
	jr	rfrme




; SP!
#IFDEF UDG
SPSTOH	.byte	$83,"SP!"
#ELSE
SPSTOH	.byte	$83,"SP",$5f
#ENDIF
	.word	RATH
SPSTO	.word	*+2
	ld	hl,DStack
	ld	(DSP),hl
	jp	NEXT




; SP@
#IFDEF UDG
SPFETH	.byte	$83,"SP@"
#ELSE
SPFETH	.byte	$83,"SP?"
#ENDIF
	.word	SPSTOH
SPFET	.word	*+2
	ld	hl,(DSP)
	call	PushOne
	jp	NEXT




; K

KDOH	.byte	$81,"K"
	.word	SPFETH
KDO	.word	*+2
	ld	hl,$0008
	jr	jdoent




; J

JDOH	.byte	$81,"J"
	.word	KDOH
JDO	.word	*+2
	ld	hl,$0004
jdoent	add	hl,sp
	ld	e,(hl)
	inc	hl
	ld	d,(hl)
	ex	de,hl
	call	PushOne
	jp	NEXT


; I

IDOH	.byte	$81,"I"
	.word	JDOH
IDO	.word	*+2
	jr	RAT+2




; U<

ULESSH	.byte	$82,"U<"
	.word	IDOH
ULESS	.word	COLON
	.word	ZERO
	.word	SWAP
	.word	ZERO
	.word	DLESS
	.word	SEMI




; U*

USTARH	.byte	$82,"U*"
	.word	ULESSH
USTAR	.word	*+2
	push	bc
	call	PopOne
	ex	de,hl
	call	PopOne
	ld	b,h
	ld	c,l
	call	M16x16
	call	PushOne
	ex	de,hl
	call	PushOne
	pop	bc
	jp	NEXT



; *

STARH	.byte	$81,"*"
	.word	USTARH
STAR	.word	COLON
	.word	USTAR
	.word	DROP
	.word	SEMI




; */

SSLASHH	.byte	$82,"*/"
	.word	STARH
SSLASH	.word	COLON
	.word	SSMOD
	.word	SWAP
	.word	DROP
	.word	SEMI




; U/MOD
USLMODH	.byte	$85,"U/MOD"
	.word	SSLASHH
USLMOD	.word	*+2
	push	bc
	call	PopOne
	ld	b,h
	ld	c,l
	call	PopOne
	ex	de,hl
	call	PopOne
	ld	a,l
	ld	(Tmp),a
	ld	a,h
	ld	(Tmp+1),a
	ex	de,hl
	call	D32x16b
	ld	l,e
	ld	h,d
	call	PushOne
	ld	a,(Tmp)
	ld	l,a
	ld	a,(Tmp+1)
	ld	h,a
	call	PushOne
	pop	bc
	jp	NEXT




; M/
MSLASHH	.byte	$82,"M/"
	.word	USLMODH
MSLASH	.word	COLON
	.word	OVER
	.word	TOR
	.word	TOR
	.word	DABS
	.word	RAT
	.word	ABS
	.word	USLMOD
	.word	RFROM
	.word	RAT
	.word	XOR
	.word	PM
	.word	SWAP
	.word	RFROM
	.word	PM
	.word	SWAP
	.word	SEMI



 
; M/MOD
MSMODH	.byte	$85,"M/MOD"
	.word	MSLASHH
MSMOD	.word	COLON
	.word	TOR
	.word	ZERO
	.word	RAT
	.word	USLMOD
	.word	RFROM
	.word	SWAP
	.word	TOR
	.word	USLMOD
	.word	RFROM
	.word	SEMI




; M*
MSTARH	.byte	$82,"M*"
	.word	MSMODH
MSTAR	.word	COLON
	.word	TDUP
	.word	XOR
	.word	TOR
	.word	ABS
	.word	SWAP
	.word	ABS
	.word	USTAR
	.word	RFROM
	.word	DPM
	.word	SEMI




; /MOD
SLMODH	.byte	$84,"/MOD"
	.word	MSTARH
SLMOD	.word	COLON
	.word	TOR
	.word	STOD
	.word	RFROM
	.word	MSLASH
	.word	SEMI




 ; /
SLASHH	.byte	$81,"/"
	.word	SLMODH
SLASH	.word	COLON
	.word	SLMOD
	.word	SWAP
	.word	DROP
	.word	SEMI




; MOD
MODH	.byte	$83,"MOD"
	.word	SLASHH
MOD	.word	COLON
	.word	SLMOD
	.word	DROP
	.word	SEMI




; */MOD
SSMODH	.byte	$85,"*/MOD"
	.word	MODH
SSMOD	.word	COLON
	.word	TOR
	.word	MSTAR
	.word	RFROM
	.word	MSLASH
	.word	SEMI




; U/ - same as U/MOD. Might incorposate a zero check at some point
USLASHH	.byte	$82,"U/"
	.word	SSMODH
USLASH	.word	COLON
	.word	USLMOD
	.word	SEMI




; +
PLUSH	.byte	$81,"+"
	.word	USLASHH
PLUS	.word	*+2
	call	PopOne
	ex	de,hl
	call	PopOne
	add	hl,de
	call	PushOne
	jp	NEXT




; -

SUBH	.byte	$81,"-"
	.word	PLUSH
SUB	.word	COLON
	.word	NEGATE
	.word	PLUS
	.word	SEMI




; D-

DMINUSH	.byte	$82,"D-"
	.word	SUBH
DMINUS	.word	COLON
	.word	DNEGAT
	.word	DPLUS
	.word	SEMI




; D+
DPLUSH	.byte	$82,"D+"
	.word	DMINUSH
DPLUS	.word	*+2
	push	bc
	call	PopOne
	ld	b,h		; msb n1
	ld	c,l

	call	PopOne
	ex	de,hl		; lsb n1
	
	call	PopOne		; msb n2
	push	hl

	call	PopOne		; lsb n2
	add	hl,de
	push	af		; Save carry

	call	PushOne

	pop	af		; restore carry
	pop	hl		; restore msb	n2
	adc	hl,bc		; add msbs with carry
	call	PushOne
	pop	bc
	jp	NEXT




; D<
	.MODULE	DLESS_M
DLESSH	.byte	$82,"D<"
	.word	DPLUSH
DLESS	.word	*+2
	call	PopOne
	push	hl		; n2 msb
	call	PopOne
	ex	de,hl		; n2 lsb in DE
	call	PopOne
	push	hl		; n1 msb
	call	PopOne		; n1 lsb in HL
	and	a
	sbc	hl,de
	pop	hl
	pop	de
	sbc	hl,de

	ld	de,$0000
	jp	po,_1
	ld	a,h
	xor	$80
_1	jp	p,_2
	inc	e
_2	ex	de,hl
	call	PushOne
	jp	NEXT




; +! (or + on the non-UDG ZX81)
#IFDEF UDG
PSTORH	.byte	$82,"+!"
#ELSE
PSTORH	.byte	$82,"+",$5f
#ENDIF
	.word	DLESSH
PSTOR	.word	*+2
	call	PopOne
	ex	de,hl
	call	PopOne
	ex	de,hl
	and	a
	ld	a,(hl)
	add	a,e
	ld	(hl),a
	inc	hl
	ld	a,(hl)
	adc	a,d
	ld	(hl),a
	jp	NEXT




; ! (or  on ZX81)
#IFDEF UDG
STOREH	.byte	$81,"!"
#ELSE
STOREH	.byte	$81,$5f
#ENDIF
.word	PSTORH
STORE	.word	*+2
	call	PopOne
	ex	de,hl
	call	PopOne
	ex	de,hl
	ld	(hl),e
	inc	hl
	ld	(hl),d
	jp	NEXT




; C! (or C on ZX81)
#IFDEF UDG
CSTOREH	.byte	$82,"C!"
#ELSE
CSTOREH	.byte	$82,"C",$5f
#ENDIF
	.word	STOREH
CSTORE	.word	*+2
	call	PopOne
	ex	de,hl
	call	PopOne
	ex	de,hl
	ld	(hl),e
	jp	NEXT




; HLD
HLDH	.byte	$83,"HLD"
	.word	CSTOREH
HLD	.word	DoUser
	.byte	0




; HOLD
HOLDH	.byte	$84,"HOLD"
	.word	HLDH
HOLD	.word	COLON
	.word	TRUE
	.word	HLD
	.word	PSTOR
	.word	HLD
	.word	AT
	.word	CSTORE
	.word	SEMI




; SIGN


SIGNH	.byte	$84,"SIGN"
	.word	HOLDH
SIGN	.word	COLON
	.word	ROT
	.word	ZLESS
	.word	ZBRANCH
sieb	.byte	siea-sieb
	.word	LIT
	.word	'-'
	.word	HOLD
siea	.word	SEMI




; CONVERT

CONVERTH
	.byte	$87,"CONVERT"
	.word	SIGNH
CONVERT	.word	COLON
conveb	.word	ONEP
	.word	DUP
	.word	TOR
	.word	CAT
	.word	BASE
	.word	CAT
	.word	DIGIT
	.word	ZBRANCH
convsa	.byte	convea-convsa
	.word	SWAP
	.word	BASE
	.word	CAT
	.word	USTAR
	.word	DROP
	.word	ROT
	.word	BASE
	.word	CAT
	.word	USTAR
	.word	DPLUS
	.word	RFROM
	.word	BRANCH
convsb	.byte	conveb-convsb
convea	.word	RFROM
	.word	SEMI




; EXPECT
	.MODULE EXPECT_M
EXPECTH	.byte	$86,"EXPECT"
	.word	CONVERTH
EXPECT	.word	*+2
	push	bc
	call	EXInput
	pop	bc

	ld	hl,(IPBadd)
	ld	a,(IPBcnt)
	and	a
	jr	z,_2
	add	a,l
	ld	l,a
	jr	nc,_1
	inc	h
_1	ld	(hl),$00
	inc	hl
	ld	(hl),$00
_2	jp	NEXT




INTEH	.byte	$89,"INTERPRET"
	.word	EXPECTH
INTE	.word	COLON

	.word	INTER
	.word	QUESTION
	.word	DROP		; Drop OK


	.word	LIT
	.word	LBuff+LBSIZE
	.word	LIT
	.word	LBP
	.word	STORE

	.word	SEMI




; QUERY

QUERYH	.byte	$85,"QUERY"
	.word	INTEH
QUERY	.word	COLON

	.word	TIB
	.word	AT
	.word	CLIT
	.byte	$7a

	.word	EXPECT

	.word	ZERO
	.word	IN
	.word	STORE

	.word	LIT
	.word	LBuff
	.word	LIT
	.word	LBP
	.word	STORE
	
	.word	SEMI




; DIGIT

	.MODULE DIGIT_M
DIGITH	.byte	$85,"DIGIT"
	.word	QUERYH
DIGIT	.word	*+2
	call	PopOne
	ex	de,hl		; E=base
	call	PopOne		; L=char
	ld	a,l
	sub	'0'		; '0'
	jr	c,_3
	cp	$0a
	jr	c,_1
	sub	$07
	cp	$0a
	jr	c,_3

_1	cp	e
	jr	nc,_3

	ld	e,a
	ld	l,$01

	ex	de,hl
	call	PushOne
	ex	de,hl
_2	call	PushOne
	jp	NEXT
_3
	ld	l,$00
	jr	_2




; ?BRK
	.MODULE QBRK_M
QBRKH	.byte	$84,"?BRK"
	.word	DIGITH
QBRK	.word	*+2
	ld	de,0
	call	BREAK-1
	jr	c,_1
	inc	de
_1	ex	de,hl
	call	PushOne
	jp	NEXT




; KEY

KEYH	.byte	$83,"KEY"
	.word	QBRKH
KEY	.word	*+2
	call	ScanRepeat
	ld	h,$00
	ld	l,a
	call	PushOne
	jp	NEXT


	

; PAD

PADH	.byte	$83,"PAD"
	.word	KEYH
PAD	.word	DoCon
	.word	PADL




; BL

BLH	.byte	$82,"BL"
	.word	PADH
BL	.word	DoCon
	.word	BLANK




; WDSZ

WDSZH	.byte	$84,"WDSZ"
	.word	BLH
WDSZ	.word	DoCon
	.word	WBSIZE




; WBFR

WBFRH	.byte	$84,"WBFR"
	.word	WDSZH
WBFR	.word	DoCon
	.word	WBuff




; OUT

OUTH	.byte	$83,"OUT"
	.word	WBFRH
OUT	.word	DoUser
	.byte	$02




; >IN

INH	.byte	$83,">IN"
	.word	OUTH
IN	.word	DoUser
	.byte	$04




; BLK

BLKH	.byte	$83,"BLK"
	.word	INH
BLK	.word	DoUser
	.byte	$06




; TIB

TIBH	.byte	$83,"TIB"
	.word	BLKH
TIB	.word	DoUser
	.byte	$08




; SCR

SCRH	.byte	$83,"SCR"
	.word	TIBH
SCR	.word	DoUser
	.byte	$0a




; WARNING

WARNH	.byte	$87,"WARNING"
	.word	SCRH
WARN	.word	DoUser
	.byte	$0c




; R#

#IFDEF UDG
RHASHH	.byte	$82,"R#"
#ELSE
RHASHH	.byte	$82,"R$"
#ENDIF
	.word	WARNH
RHASH	.word	DoUser
	.byte	$0e




; PAGE

CLSH	.byte	$84,"PAGE"
	.word	RHASHH
CLRS	.word	*+2
	call	Clear
	jp	NEXT




; CR

CRH	.byte	$82,"CR"
	.word	CLSH
CR	.word	*+2
	call	CRLF
	jp	NEXT




; EMIT
EMITH	.byte	$84,"EMIT"
	.word	CRH
EMIT	.word	*+2
	ld	hl,Out
	inc	(hl)
	call	PopOne
	ld	a,l
;	and	$7f
	call	PrtChr
	jp	NEXT




; SPACE
SPACEH	.byte	$85,"SPACE"
	.word	EMITH
SPACE	.word	COLON
	.word	BL
	.word	EMIT
	.word	SEMI




; SPACES

SPACESH	.byte	$86,"SPACES"
	.word	SPACEH
SPACES	.word	COLON
	.word	ZERO
	.word	MAX
	.word	QDUP
	.word	ZBRANCH
spcsa	.byte	spcea-spcsa
	.word	ZERO
	.word	S_DO
spcsb	.word	SPACE
	.word	S_LOOP
spceb	.byte	spcsb-spceb
spcea	.word	SEMI




	.MODULE TYPE_M
TYPEH	.byte	$84,"TYPE"
	.word	SPACESH
TYPE	.word	COLON
	.word	DUP
	.word	ZGTR
	.word	ZBRANCH
typsa	.byte	typea-typsa
	.word	OVER
	.word	PLUS
	.word	SWAP
	.word	S_DO
typeb	.word	IDO
	.word	CAT
	.word	EMIT
	.word	S_LOOP
typsb	.byte	typeb-typsb
	.word	BRANCH
typsc	.byte	typec-typsc
typea	.word	TDROP
typec	.word	SEMI




; D.R

DDOTRH	.byte	$83,"D.R"
	.word	TYPEH
DDOTR	.word	COLON
	.word	TOR
	.word	SWAP
	.word	OVER
	.word	DABS
	.word	BDIG
	.word	DIGS
	.word	SIGN
	.word	EDIG
	.word	RFROM
	.word	OVER
	.word	SUB
	.word	SPACES
	.word	TYPE
	.word	SEMI




; D.

DDOTH	.byte	$82,"D."
	.word	DDOTRH
DDOT	.word	COLON
	.word	ZERO
	.word	DDOTR
	.word	SPACE
	.word	SEMI




; .R

DOTRH	.byte	$82,".R"
	.word	DDOTH
DOTR	.word	COLON
	.word	TOR
	.word	STOD
	.word	RFROM
	.word	DDOTR
	.word	SEMI




; .

DOTH	.byte	$81,"."
	.word	DOTRH
DOT	.word	COLON
	.word	STOD
	.word	DDOT
	.word	SEMI



; U.

UDOTH	.byte	$82,"U."
	.word	DOTH
UDOT	.word	COLON
	.word	ZERO
	.word	DDOT
	.word	SEMI




; ? (?? on non-UDG ZX81)
#IFDEF UDG
QUESTH	.byte	$81,"?"
#ELSE
QUESTH	.byte	$82,"??"
#ENDIF
	.word	UDOTH
QUEST	.word	COLON
	.word	AT
	.word	DOT
	.word	SEMI




; DEC.

DECDOTH	.byte	$84,"DEC."
	.word	QUESTH
DECDOT	.word	COLON
	.word	BASE
	.word	AT
	.word	SWAP
	.word	DECIM
	.word	DOT
	.word	BASE
	.word	STORE
	.word	SEMI



; H.

HEXDOTH	.byte	$82,"H."
	.word	DECDOTH
HEXDOT	.word	COLON
	.word	BASE
	.word	AT
	.word	SWAP
	.word	HEX
	.word	DOT
	.word	BASE
	.word	STORE
	.word	SEMI




DECIMH	.byte	$87,"DECIMAL"
	.word	HEXDOTH
DECIM	.word	*+2
	ld	hl,Base
	ld	(hl),$0a
	jp	NEXT



; Set base to hexadecimal

HEXH	.byte	$83,"HEX"
	.word	DECIMH
HEX	.word	*+2
	ld	hl,Base
	ld	(hl),$10
	jp	NEXT




; MSG#
#IFDEF UDG
MSGNUMH	.byte	$84,"MSG#"
#ELSE
MSGNUMH	.byte	$84,"MSG$"
#ENDIF
	.word	HEXH
MSGNUM	.word	COLON
	.word	QDUP
	.word	ZBRANCH
msgnsa	.byte	-msgnsa
	.word	BDOTQ
#IFDEF	UDG
	.byte	$06,"MSG # "
#ELSE
	.byte	$06,"MSG $ "
#ENDIF
	.word	DECDOT
msgnea	.word	SEMI




; MESSAGE

MESSH	.byte	$87,"MESSAGE"
	.word	MSGNUMH
MESS	.word	COLON
	.word	DUP
	.word	DUP
	.word	ZGTR
	.word	SWAP
	.word	CLIT
	.byte	10		; AJB - allow 10 error messages
	.word	LESS
	.word	AND
	.word	ZBRANCH
messsa	.byte	messea-messsa
	.word	LIT
	.word	Messages
	.word	SWAP
	.word	ZERO
	.word	S_DO
messda	.word	DUP
	.word	CAT
	.word	PLUS
	.word	ONEP
	.word	S_LOOP
messdb	.byte	messda-messdb	
	.word	COUNT
	.word	TYPE
	.word	BRANCH
messsb	.byte	messeb-messsb
messea	.word	MSGNUM
messeb	.word	SEMI



Messages
	.word	1
	.byte	14,"EXECUTION ONLY"
	.byte	10,"OFF SCREEN"
	.byte	16,"COMPILATION ONLY"
	.byte	9,"UNDEFINED"
	.byte	7,"NO ROOM"




; ERROR

ERRORH	.byte	$85,"ERROR"
	.word	MESSH
ERROR	.word	COLON
	.word	WARN
	.word	AT
	.word	ZLESS
	.word	ZBRANCH
errsa	.byte	errea-errsa
	.word	ABORT
	.word	BRANCH
errsb	.byte	erreb-errsb
errea	.word	MESS
	.word	SPSTO
	.word	QUESTION
erreb	.word	SEMI




; ?ERROR

QERRH	.byte	$86,"?ERROR"
	.word	ERRORH
QERR	.word	COLON
	.word	SWAP
	.word	ZBRANCH
qerrsa	.byte	qerrea-qerrsa
	.word	ERROR
	.word	BRANCH
qerrsb	.byte	qerreb-qerrsb
qerrea	.word	DROP
qerreb	.word	SEMI




; S->D - Change signed single precision number to signed double
STODH	.byte	$84,"S->D"
	.word	QERRH
STOD	.word	*+2
	call	PopOne
	ld	de,$0000
	bit	7,h
	jr	z,_stodp
	dec	de
_stodp	call	PushOne
	ex	de,hl
	call	PushOne
	jp	NEXT




; D32x16 divide: HLDE / BC - quotient BC, remainder HL
D32x16	.MODULE D3216_M
	ld	a,$10
	sla	e
	rl	d
_1	adc	hl,hl
	jr	nc,_2
	or	a
	sbc	hl,bc
	or	a
	jr	_3
_2	sbc	hl,bc
	jr	nc,_3
	add	hl,bc
	scf
_3	rl	e
	rl	d
	dec	a
	jr	nz,_1

	ld	a,d
	cpl
	ld	b,a
	ld	a,e
	cpl
	ld	c,a
	ret




; HL(Tmp)/BC -> HL(Tmp) quotient,  DE remainder

D32x16b 
	ex	de,hl
	xor	a
	ld	h,a
	ld	l,a
	sub	c
	ld	c,a
	sbc	a,a
	sub	b
	ld	b,a

	ld	a,d
	call	d3216_s8
	rla
	ld	d,a

	ld	a,e
	call	d3216_s8
	rla
	ld	e,a

	ld	a,(Tmp+1)
	call	d3216_s8
	rla
	ld	(Tmp+1),a

	ld	a,(Tmp)
	call	d3216_s8
	rla
	ld	(Tmp),a

	ex	de,hl
	ret

d3216_s8
	call	_10
_10	call	_20
_20	call	d3216_s	
d3216_s	rla
	adc	hl,hl
	jr	c,_30
	add	hl,bc
	ret	c
	sbc	hl,bc
	ret
_30	add	hl,bc
	scf
	ret




; 16x16 multiply: BC*DE result in DE:HL
M16x16	.MODULE m1616_M
	ld	hl,0
	sla	e
	rl	d
	jr	nc,_1
	ld	h,b
	ld	l,c
_1	ld	a,$0f
_2	add	hl,hl
	rl	e
	rl	d
	jr	nc,_3
	add	hl,bc
	jr	nc,_3
	inc	de
_3	dec	a
	jr	nz,_2
	ret





;  <# (or <$ for non-UDG ZX81)
#IFDEF UDG
BDIGH	.byte	$82,"<#"
#ELSE
BDIGH	.byte	$82,"<$"
#ENDIF
	.word	STODH
BDIG	.word	COLON
	.word	PAD
	.word	HLD
	.word	STORE
	.word	SEMI




; # (or $ for non-UDG ZX81)
#IFDEF UDG
DIGH	.byte	$81,"#"
#ELSE
DIGH	.byte	$81,"$"
#ENDIF

	.word	BDIGH
DIG	.word	COLON
	.word	BASE
	.word	CAT
	.word	MSMOD
	.word	ROT
	.word	LIT
	.word	$09
	.word	OVER
	.word	LESS
	.word	ZBRANCH
digsa	.byte	digea-digsa
	.word	LIT
	.word	$07
	.word	PLUS
digea	.word	LIT
	.word	'0'
	.word	PLUS
	.word	HOLD
	.word	SEMI




; #S (or $S on the non-UDG ZX81)
#IFDEF UDG
DIGSH	.byte	$82,"#S"
#ELSE
DIGSH	.byte	$82,"$S"
#ENDIF
	.word	DIGH
DIGS	.word	COLON
dgse	.word	DIG
	.word	TDUP
	.word	OR
	.word	ZEROEQ
	.word	ZBRANCH
dgsa	.byte	dgse-dgsa
	.word	SEMI




;  #> (or $> for non-UDG ZX81)
#IFDEF UDG
EDIGH	.byte	$82,"#>"
#ELSE
EDIGH	.byte	$82,"$>"
#ENDIF
	.word	DIGSH
EDIG	.word	COLON
	.word	TDROP
	.word	HLD
	.word	AT
	.word	PAD
	.word	OVER
	.word	SUB
	.word	SEMI




;  HERE

HEREH	.byte	$84,"HERE"
	.word	EDIGH
HERE	.word	*+2
	ld	hl,(DP)
	call	PushOne
	jp	NEXT



; Fix up the NFA/LFA & CFA areas of a new word after its token has been
; Added to the top of the dictionary by CREATE. It expects
; the NFA of the previous word to be on the stack.

	.MODULE MKHEAD_M
MKHEAD	.word	*+2
	ld	hl,(DP)
	ld	a,(hl)
	push	af
	or	$80
	ld	(hl),a
	pop	af

	add	a,l
	ld	l,a
	jr	nc,_1
	inc	h
_1	ld	a,(hl)

	inc	hl		; HL=Link address

	ex	de,hl
	call	PopOne
	ex	de,hl
	ld	(hl),e
	inc	hl
	ld	(hl),d
	inc	hl		; HL= Code address

	ld	de,DoVar

	ld	(hl),e		; Store DoVar address
	inc	hl
	ld	(hl),d
	inc	hl
	ld	(DP),hl		; and update DP
	jp	NEXT




; CFA - Convert PFA to CFA

CFAH	.byte	$83,"CFA"
	.word	HEREH
CFA	.word	*+2
	call	PopOne
	dec	hl
	dec	hl
	call	PushOne
	jp	NEXT



 
; PFA - Convert NFA to PFA
	.MODULE PFA_M
PFAH	.byte	$83,"PFA"
	.word	CFAH
PFA	.word	*+2
	call	PopOne
	ld	a,(hl)
	and	$3f
	add	a,l
	ld	l,a
	jr	nc,_1
	inc	h
_1	inc	hl
	inc	hl
	inc	hl
	inc	hl
	inc	hl
	call	PushOne
	jp	NEXT




; NFA - Convert Parameter filed address to NFA
	.MODULE NFA_M
NFAH	.byte	$83,"NFA"
	.word	PFAH
NFA	.word	*+2
	call	PopOne
	dec	hl
	dec	hl
	dec	hl
	dec	hl
	dec	hl

_1	dec	hl
	bit	7,(hl)
	jr	z,_1
	
	call	PushOne
	jp	NEXT




; LFA - Convert parameter field address to LFA

LFAH	.byte	$83,"LFA"
	.word	NFAH
LFA	.word	*+2
	call	PopOne
	dec	hl
	dec	hl
	dec	hl
	dec	hl
	call	PushOne
	jp	NEXT




; DLITERAL

DLITERH	.byte	$88,"DLITERAL"
	.word	LFAH
DLITER	.word	COLON
	.word	STATE
	.word	CAT
	.word	ZBRANCH
dlitesa	.byte	dliteea-dlitesa
	.word	SWAP
	.word	LITER
	.word	LITER
dliteea	.word	SEMI




; LITERAL

LITERH	.byte	$87,"LITERAL"
	.word	DLITERH
LITER	.word	COLON
	.word	STATE
	.word	CAT
	.word	ZBRANCH
litesa	.byte	liteea-litesa
	.word	COMP
	.word	LIT
	.word	COMMA
liteea	.word	SEMI




; DO,

DOCOMMH	.byte	$83,"DO,"
	.word	LITERH
DOCOMM	.word	COLON
	.word	COMMA
	.word	HERE
	.word	SEMI



; S_LEAVE - Copy second stack item over first to cause loop termination
; at the next iteration

S_LEAVE	.word	*+2
	pop	hl
	pop	hl
	push	hl
	push	hl
	jp	NEXT



; LEAVE

LEAVEH	.byte	$c5,"LEAVE"
	.word	DOCOMMH
LEAVE	.word	COLON
	.word	QCOMP
	.word	LIT
	.word	S_LEAVE
	.word	COMMA
	.word	SEMI




; +LOOP

PLOOPH	.byte	$c5,"+LOOP"
	.word	LEAVEH
PLOOP	.word	COLON
	.word	QCOMP
	.word	LIT
	.word	S_XLOOP
	.word	COMMA
	.word	HERE
	.word	SUB
	.word	CCOMMA
	.word	SEMI




; LOOP

LOOPH	.byte	$c4,"LOOP"
	.word	PLOOPH
LOOP	.word	COLON
	.word	QCOMP
	.word	LIT
	.word	S_LOOP
	.word	COMMA
	.word	HERE
	.word	SUB
	.word	CCOMMA
	.word	SEMI




; DO

DOH	.byte	$c2,"DO"
	.word	LOOPH
DO	.word	COLON
	.word	QCOMP
	.word	LIT
	.word	S_DO
	.word	DOCOMM
	.word	SEMI




; ALLOT

ALLOTH	.byte	$85,"ALLOT"
	.word	DOH
ALLOT	.word	*+2
	ld	de,(DP)
	call	PopOne
	add	hl,de
	ld	(DP),hl
	jp	NEXT




; DEFINITIONS

DEFSH	.byte	$8b,"DEFINITIONS"
	.word	ALLOTH
DEFS	.word	COLON
	.word	CONTEXT
	.word	AT
	.word	CURRENT
	.word	STORE
	.word	EXIT




; TOGGLE

TOGGLEH	.byte	$86,"TOGGLE"
	.word	DEFSH
TOGGLE	.word	*+2
	call	PopOne
	ex	de,hl
	call	PopOne
	ld	a,(hl)
	xor	e
	ld	(hl),a
	jp	NEXT




; IMMEDIATE

IMMEDH	.byte	$89,"IMMEDIATE"
	.word	TOGGLEH
IMMED	.word	COLON
	.word	LAST
	.word	LIT
	.word	$40
	.word	TOGGLE
	.word	SEMI




; [ - '(' on non-UDG ZX81
#IFDEF UDG
LBRACEH	.byte	$c1,"["
#ELSE
LBRACEH	.byte	$c1,"("
#ENDIF
	.word	IMMEDH
LBRACE	.word	COLON
	.word	ZERO
	.word	STATE
	.word	CSTORE
	.word	SEMI




; ] - ')' on non-UDG ZX81
#IFDEF UDG
RBRACEH	.byte	$81,"]"
#ELSE
RBRACEH	.byte	$81,")"
#ENDIF
	.word	LBRACEH
RBRACE	.word	COLON
	.word	ONE
	.word	STATE
	.word	CSTORE
	.word	SEMI




; [COMPILE] - (COMPILE) on ZX81
#IFDEF UDG
BCOMPH	.byte	$89,"[COMPILE]"
#ELSE
BCOMPH	.byte	$89,"(COMPILE)"
#ENDIF
	.word	RBRACEH
BCOMP	.word	COLON
	.word	FIND
	.word	DUP
	.word	ZEROEQ
	.word	ZBRANCH
bcomsa	.byte	bcomsa-bcomea
	.word	QUESTION
bcomea	.word	COMMA
	.word	SEMI




; BACK

BACKH	.byte	$84,"BACK"
	.word	BCOMPH
BACK	.word	COLON
	.word	HERE
	.word	SUB
	.word	CCOMMA
	.word	SEMI




; ?EXEC

QEXECH	.byte	$85,"?EXEC"
	.word	BACKH
QEXEC	.word	COLON
	.word	STATE
	.word	CAT
	.word	CLIT
	.byte	$01
	.word	QERR
	.word	SEMI




; ?COMP

QCOMPH	.byte	$85,"?COMP"
	.word	QEXECH
QCOMP	.word	COLON
	.word	STATE
	.word	CAT
	.word	ZEROEQ
	.word	CLIT
	.byte	$03
	.word	QERR
	.word	SEMI




; Store the CFA of the following word in the dictionary and skip
; that word.

; COMPILE

COMPH	.byte	$87,"COMPILE"
	.word	QCOMPH
COMP	.word	COLON
	.word	QCOMP
	.word	RFROM
	.word	DUP
	.word	TWOP
	.word	TOR
	.word	AT
	.word	COMMA
	.word	SEMI




; UNTIL

UNTILH	.byte	$c5,"UNTIL"
	.word	COMPH
UNTIL	.word	COLON
	.word	QCOMP
	.word	LIT
	.word	ZBRANCH
	.word	DOCOMM
	.word	SUB
	.word	CCOMMA
	.word	SEMI



; REPEAT

REPEATH	.byte	$c6,"REPEAT"
	.word	UNTILH
REPEAT	.word	COLON
	.word	QCOMP
	.word	SWAP
	.word	AGAIN
	.word	THEN
	.word	SEMI




; WHILE

WHILEH	.byte	$c5,"WHILE"
	.word	REPEATH
WHILE	.word	COLON
	.word	QCOMP
	.word	IF
	.word	SEMI




; AGAIN

AGAINH	.byte	$c5,"AGAIN"
	.word	WHILEH
AGAIN	.word	COLON
	.word	QCOMP
	.word	LIT
	.word	BRANCH
	.word	DOCOMM
	.word	SUB
	.word	CCOMMA
	.word	SEMI
	



; BEGIN

BEGINH	.byte	$c5,"BEGIN"
	.word	AGAINH
BEGIN	.word	COLON
	.word	QCOMP
	.word	HERE
	.word	SEMI




; THEN

THENH	.byte	$c4,"THEN"
	.word	BEGINH
THEN	.word	COLON
	.word	QCOMP
	.word	HERE
	.word	OVER
	.word	SUB
	.word	SWAP
	.word	CSTORE
	.word	SEMI



; ELSE

ELSEH	.byte	$c4,"ELSE"
	.word	THENH
ELSE	.word	COLON
	.word	QCOMP
	.word	LIT
	.word	BRANCH
	.word	DOCOMM
	.word	ZERO
	.word	CCOMMA
	.word	SWAP
	.word	THEN
	.word	SEMI




; IF

IFH	.byte	$c2,"IF"
	.word	ELSEH
IF	.word	COLON
	.word	QCOMP
	.word	LIT
	.word	ZBRANCH
	.word	DOCOMM
	.word	ZERO
	.word	CCOMMA
	.word	SEMI




; NOOP

NOOPH	.byte	$84,"NOOP"
	.word	IFH
NOOP	.word	COLON
	.word	SEMI




; COUNT

COUNTH	.byte	$85,"COUNT"
	.word	NOOPH
COUNT	.word	*+2
	call	PopOne
	ld	e,(hl)
	ld	d,$00
	inc	hl
	call	PushOne
	ex	de,hl
	call	PushOne
	jp	NEXT




; (WORD)

PWORDH	.byte	$86,"(WORD)"
	.word	COUNTH
PWORD	.word	*+2
	push	bc
	call	TokenParse	; Get position of string
	call	PushOne		; Start address
	ld	h,0		; and count
	ld	l,b
	call	PushOne
	pop	bc
	jp	NEXT




; WORD

	.MODULE WORD_M
WORDH	.byte	$84,"WORD"
	.word	PWORDH
WORD	.word	*+2
	call	TokenMC
	ld	hl,(DP)
	ld	de,WBuff

	xor	a		; Assume no token
	ld	(de),a

	ld	a,(hl)
	cp	$01
	jr	nz,_1

	inc	hl
	ld	a,(hl)
	dec	hl
	cp	$80
	jr	nz,_1
	ld	hl,(LBP)	; If end of line buffer then backstep LBP
	dec	hl		; so the terminator is correctly detected
	ld	(LBP),hl	; by the outer interpreter.
	jr	_3


_1	ld	a,(hl)		; Copy count byte
	ld	(de),a
	inc	hl
	inc	de
	push	bc
	ld	b,a
_2	ld	a,(hl)
	ld	(de),a
	inc	hl
	inc	de
	djnz	_2
	pop	bc

_3	ld	hl,WBuff
	call	PushOne
	jp	NEXT




; TEXT,
	.MODULE TXTCOM_M
TXTCOMH	.byte	$85,"TEXT,"
	.word	WORDH
TXTCOM	.word	*+2
	call	PopOne
	ld	a,l
	call	PopOne
	and	a
	jr	nz,_1
	ld	hl,(DP)
	ld	(hl),0
	inc	hl
	ld	(DP),hl
	jp	NEXT

_1	push	bc
	ld	b,a
	ld	de,(DP)
	ld	(de),a
	inc	de

_2	ld	a,(hl)
	ld	(de),a
	inc	hl
	inc	de
	djnz	_2

	ld	(DP),de
	pop	bc
	jp	NEXT




; String print handler
	.MODULE S_DOTQ_M
S_DOTQ	.word	*+2
	ld	a,(bc)
	ld	e,a
_1	inc	bc
	ld	a,(bc)
	call	PrtChr
	dec	e
	jr	nz,_1
	inc	bc
	jp	NEXT




; (.")

BDOTQH	.byte	$84,"(.",$22,")"
	.word	TXTCOMH
BDOTQ	.word	COLON
	.word	RAT
	.word	COUNT
	.word	DUP
	.word	ONEP
	.word	RFROM
	.word	PLUS
	.word	TOR
	.word	TYPE
	.word	SEMI




; ."

DOTQH	.byte	$c2,".",$22
	.word	BDOTQH
DOTQ	.word	COLON
	.word	COMP
	.word	S_DOTQ
	.word	LIT
	.word	$22
	.word	WORD
	.word	COUNT
	.word	TXTCOM
	.word	SEMI




; ;

SEMIH	.byte	$c1,";"
	.word	DOTQH
SEM	.word	COLON
	.word	QCOMP
	.word	LIT
	.word	SEMI
	.word	COMMA
	.word	ZERO
	.word	STATE
	.word	CSTORE
	.word	SEMI




; :

COLONH	.byte	$c1,":"
	.word	SEMIH
COLO	.word	COLON
	.word	QEXEC
	.word	CURRENT
	.word	AT
	.word	CONTEXT
	.word	STORE
	.word	CREATE
	.word	LIT
	.word	COLON
	.word	LAST
	.word	PFA
	.word	CFA
	.word	STORE
	.word	ONE
	.word	STATE
	.word	CSTORE
	.word	SEMI
	

DoDoes
	ex	de,hl
	
	call	PushOne
	pop	de
	jp	COLON




; DOES>

DOESH	.byte	$c5,"DOES>"
	.word	COLONH
DOES	.word	COLON
	.word	COMP
	.word	SCODE
	.word	LIT
	.word	$cd
	.word	CCOMMA
	.word	COMP
	.word	DoDoes
	.word	SEMI




; CREATE

CREATEH	.byte	$86,"CREATE"
	.word	DOESH
CREATE	.word	COLON
	.word	LAST
	.word	BL
	.word	TOKEN
	.word	HERE
	.word	CURRENT
	.word	AT
	.word	STORE
	.word	MKHEAD
	.word	SEMI




; ERASE

ERASEH	.byte	$85,"ERASE"
	.word	CREATEH
ERASE	.word	COLON
	.word	ZERO
	.word	FILL
	.word	SEMI




; BLANKS

BLANKSH	.byte	$86,"BLANKS"
	.word	ERASEH
BLANKS	.word	COLON
	.word	BL
	.word	FILL
	.word	SEMI




; USE

USEH
	.byte	$83,"USE"
	.word	BLANKSH
USE	.word	DoVar
Use	.word	$00




; PREV

PREVH
	.byte	$84,"PREV"
	.word	USEH
PREV	.word	DoVar
Prev	.word	$00




; LIMIT

LIMITH	.byte	$85,"LIMIT"
	.word	PREVH
LIMIT	.word	DoCon
	.word	BuffEnd




; FIRST

FIRSTH	.byte	$85,"FIRST"
	.word	LIMITH
FIRST	.word	DoCon
	.word	Buff1




; NOVEC

NOVECH	.byte	$85,"NOVEC"
	.word	FIRSTH
NOVEC	.word	COLON
	.word	CLIT
	.byte	4
	.word	ERROR
	.word	SEMI



; EXVEC:

EXVECH	.byte	$86,"EXVEC:"
	.word	NOVECH
EXVEC	.word	COLON
	.word	CREATE
	.word	LIT
	.word	NOVEC
	.word	COMMA
	.word	SCODE
DOXVEC	.byte	$cd,DoDoes%256,DoDoes/256
	.word	AT
	.word	EXECUTE
	.word	SEMI



; ASSIGN

ASSIGNH	.byte	$86,"ASSIGN"
	.word	EXVECH
ASSIGN	.word	COLON
	.word	TICK
	.word	SEMI




; DOVEC

DOVECH	.byte	$85,"DOVEC"
	.word	ASSIGNH
DOVEC	.word	COLON
	.word	CFA
	.word	SWAP
	.word	STORE
	.word	SEMI



; TO-DO

TODOH	.byte	$85,"TO-DO"
	.word	DOVECH
TODO	.word	COLON
	.word	TICK
	.word	STATE
	.word	AT
	.word	ZBRANCH
tdsa	.byte	tdea-tdsa
	.word	COMP
tdea	.word	DOVEC
	.word	SEMI




; (UPDATE)

BRAUPH	.byte	$88,"(UPDATE)"
	.word	TODOH
BRAUP	.word	COLON
	.word	PREV
	.word	AT
	.word	AT
	.word	LIT
	.word	$8000
	.word	OR
	.word	PREV
	.word	AT
	.word	STORE
	.word	SEMI




; UPDATE

UPDATEH	.byte	$86,"UPDATE"
	.word	BRAUPH
UPDATE	.word	DOXVEC
	.word	NOOP




; BUFSZ

BUFSZH	.byte	$85,"BUFSZ"
	.word	UPDATEH
BUFSZ	.word	DoCon
	.word	$404




; B/SCR

BSSCRH	.byte	$85,"B/SCR"
	.word	BUFSZH
BSSCR	.word	DoCon
	.word	1




; B/BUF

BSBUFH	.byte	$85,"B/BUF"
	.word	BSSCRH
BSBUF	.word	DoCon
	.word	BLKSIZE




; C/L

CSLH	.byte	$83,"C/L"
	.word	BSBUFH
CSL	.word	DoCon
	.word	$1c



; (LINE)

BRALINH	.byte	$86,"(LINE)"
	.word	CSLH
BRALINE	.word	COLON
	.word	TOR
	.word	CSL
	.word	BSBUF
	.word	SSMOD
	.word	RFROM
	.word	BSSCR
	.word	STAR
	.word	PLUS
	.word	BLOCK
	.word	PLUS
	.word	CSL
	.word	SEMI




; Does all the work for the header BUFFER

Do_Buff	.MODULE	BUFFER_M
	ld	hl,(Use)
	ld	(Tmp),hl
	ld	(Prev),hl
	ld	e,(hl)		; Screen number
	inc	hl
	ld	d,(hl)		; MSByte
	bit	7,d		; Need to save out?
	jr	z,_1		; No.

	call	TapeSave

_1	ld	hl,(Use)	; Point to next available buffer
	call	Nextbuff
	ld	(Use),hl

	call	PopOne		; Get new screen number
	ld	a,l
	ld	hl,(Prev)	; Get current buffer address

	ld	(hl),a		; Store screen number
	inc	hl
	ld	(hl),0
	inc	hl

	call	PushOne
	ret




	.MODULE TapeSave_M
TapeSave
	push	bc
	ld	a,e		; Set SCRxx text
	call	MkFname

	call	CRLF
	ld	hl,Press	; Prompt
	call	CTPrt
	ld	hl,SaveTxt
	call	CTPrt

	ld	hl,AFname
	call	CTPrt


	call	ScanRepeat

	call	SET_FAST
	
	ld	hl,(Tmp)	; Start of buffer
	push	hl
	ld	bc,BUFSIZE-3	; Make BC last byte address
	add	hl,bc
	ld	b,h
	ld	c,l
	pop	de		; Retrieve buffer start
	inc	de
	ld	a,(de)		; Reset update flag
	and	$7f
	ld	(de),a
	inc	de		; DE = First save address
	ld	hl,Fname	; HL = Name
	call	Blk_Save

	call	SLOW
	pop	bc
	ret




;  -->
	.MODULE Arrow_M
ARROWH	.byte	$83,"-->"
	.word	BRALINH
ARROW	.word	COLON
	.word	Do_Arrow
	.word	UPDBLK
	.word	INTE
	.word	SEMI




Do_Arrow
	.word	*+2
	push	bc
	ld	a,(Scnt)	; Get current input block (initially from
	inc	a		; LOAD) and increment to get next screen.
	ld	(Scnt),a	; Save in case another --> is found.

	call	Findbuff	; Screen in memory
	jr	nz,_1		; No
	inc	hl		; Yes, so point to data area
	inc	hl
	jr	_arrint		; and adjust line input pointer

_1	ld	h,0		; Put new buffer number on stack for
	ld	l,a		; Do_Buff
	call	PushOne

	call	Do_Buff		; This will stack the buffer address
				; HL=buff data addr, A=buffno
	call	PopOne		; Discard address

	call	MkFname		; Set SCRxx text
	call	BlkLd


	ld	hl,(Prev)
	inc	hl
	inc	hl
_arrint
	ld	(LBP),hl

	pop	bc
	jp	NEXT




; UPDBLK - update storage block number accoding to its address

UPDBLK	.MODULE UPDBLK_M
	.word	*+2
	push	bc
	ld 	de,$0001
	ld	bc,Buff1
	ld	hl,(Prev)
	and	a
	sbc	hl,bc
	jr	z,_1
	inc	de
_1	ld	(Blk),de
	pop	bc
	jp	NEXT


; BUFFER

BUFFERH	.byte	$86,"BUFFER"
	.word	ARROWH
BUFFER	.word	*+2
	call	Do_Buff
	jp	NEXT




; Does all the work for the header BLOCK

Do_Block .MODULE Block_M
	push	hl
	ld	a,l
	call	Findbuff
	jr	nz,_1
	inc	hl
	inc	hl
	call	PushOne
	pop	hl
	ret

_1	pop	hl
	push	hl

	call	PushOne
	call	Do_Buff		; This will stack the buffer address

	pop	hl
	ld	a,l		; Set SCRxx text
	call	MkFname

	call	CRLF
	ld	hl,Press	; Prompt
	call	CTPrt
	ld	hl,LoadTxt
	call	CTPrt
	ld	hl,AFname
	call	CTPrt

	call	ScanKBD		; Get any key press

BlkLd	call	SET_FAST

	ld	hl,(Prev)	; Start of buffer
	push	hl
	ld	bc,BUFSIZE-3	; Make BC last byte address
	add	hl,bc
	ld	b,h
	ld	c,l
	pop	de		; Retrieve buffer start
	inc	de
	ld	a,(de)		; Reset update byte
	and	$7f
	ld	(de),a
	inc	de		; DE = First save address
	ld	hl,Fname	; HL = Name

	call	Blk_Load

	call	SLOW

	ret



; BLOCK

BLOCKH	.byte	$85,"BLOCK"
	.word	BUFFERH
BLOCK	.word	*+2
	call	PopOne
	push	bc
	call	Do_Block
	pop	bc
	jp	NEXT




; LOAD

LOADH	.byte	$84,"LOAD"
	.word	BLOCKH
LOAD	.word	COLON

	.word	DUP		; For -->

	.word	LIT
	.word	Scnt
	.word	STORE		; Read block and set line buffer ptr

	.word	BLOCK
	.word	LIT
	.word	LBP
	.word	STORE

	.word	UPDBLK
	
	.word	INTE		; Interpret

	.word	ZERO
	.word	BLK
	.word	STORE

	.word	SEMI




; LIST
	.MODULE List_M
LISTH	.byte	$84,"LIST"
	.word	LOADH
LIST	.word	*+2
	push	bc

	call	PopOne
	push	hl

	call	Do_Block
	call	PopOne

	call	CRLF
	ld	hl,Lname
	call	CTPrt
	
	pop	hl
	ld	(Scr),hl


	ld	a,l
	push	af
	call	DecPrt

	ld	a,BLANK
	call 	PrtChr
	pop	af

	push	af

	call	Bin2Hex
	ld	a,d
	call	PrtChr
	ld	a,e
	call	PrtChr
	call	CRLF

	pop	af		; Locate screen data from screen number
	call	Findbuff
	inc	hl
	inc	hl
	
	xor	a
	ld	(Ltmp),a

_1	cp	LLINES
	jr	z,_4

	cp	$16		; Pause listing after a screenful
	jr	nz,_2
	push	af
	call	ScanRepeat
	pop	af

_2	call	DecPrt
	ld	a,BLANK
	call	PrtChr
	ld	a,BLANK
	call	PrtChr
	ld	b,CPL
_3	ld	a,(hl)
	call	PrtChr
	inc	hl
	djnz	_3
	ld	a,(Ltmp)
	inc	a
	ld	(Ltmp),a
	jr	_1

_4	pop	bc
	jp	NEXT




; FLUSH

FLUSHH	.byte	$85,"FLUSH"
	.word	LISTH
FLUSH	.word	COLON
	.word	SAVBUF
	.word	EMPTY
	.word	SEMI




; SAVE-BUFFERS

SAVBUFH	.byte	$8c,"SAVE-BUFFERS"
	.word	FLUSHH
SAVBUF	.word	*+2
	ld	hl,Buff1
	call	Do_SaveBuff
	ld	hl,Buff2
	call	Do_SaveBuff
	jp	NEXT



Do_SaveBuff
	ld	(Tmp),hl
	ld	e,(hl)		; Screen number
	inc	hl
	ld	d,(hl)
	bit	7,d		; Updated?
	ret	z		; No, so exit
	call	TapeSave
	ret




; EMPTY-BUFFERS

EMPTYH	.byte 	$8d,"EMPTY-BUFFERS"
	.word	SAVBUFH
EMPTY	.word	*+2
	call	EmpBufs
	jp	NEXT



; SMUDGE

SMUDGEH	.byte	$86,"SMUDGE"
	.word	EMPTYH
SMUDGE	.word	COLON
	.word	LAST
	.word	CLIT
	.byte	$20
	.word	TOGGLE
	.word	SEMI




TEST	.word	*+2
	jp	NEXT







TopNfa


; PSAVE

PSAVEH	.byte	$85,"PSAVE"
	.word	SMUDGEH
PSAVE	.word	*+2
	call	Clear
	ld	hl,BegMsg
	call	CTPrt
	call	CRLF

	ld	hl,(DP)
	ld	(E_LINE),hl
	exx
	call	SET_FAST
	ld	de,PsvTxt	; FIF filename
	call	SAVE+5
	call	SLOW
	exx
	jp	NEXT




Blk_S	.word	$00		; Start of block to Save/Load
Blk_E	.word	$00		; End of block to Save/Load



; BlkSave - Save a block of memory to cassette
; Enter: HL = Start address of name (terminated by a byte with bit 7 set)
;        DE = Start address of block to save
;	 BC = End address of block to save
;
; Exit: NC = Break pressed

Blk_Save	.MODULE BlkSvLd
	ld	(Blk_S),de
	ld	(Blk_E),bc

	ld	de,$12cb	; 5 seconds timing value
_1	call	BREAK-1		; Outer delay loop. No carry if SPACE
				; not pressed
	ret	nc

	ld	b,$00		; Inner delay loop
	djnz	*	
	dec	de		; decrement outer loop counter
	ld	a,d
	or	e
	jr	nz,_1		; and jump back if not zero.

	; Save name
_namout	call	_OutByte	; Output character
	ret	nc		; Break pressed
	bit	7,(hl)		; Test for terminator
	inc	hl		; Increment name pointer
	jr	z,_namout

	; Save data
	ld	hl,(Blk_S)
_datout	call	_OutByte
	ret	nc
	call	_BlkUpd		; Test for end of block
	jr	nc,_datout
	ret



; Output character in (hl) to cassette
; Returns: NC=Break pressed
_OutByte
	ld	e,(hl)		; Get byte to save
	scf			; Set carry as a marker
_OBbit	rl	e		; Rotate bit into carry
	jr	nz,_dobit	; Return whan marker bit has passed through.
	scf			; Carry set denotes success.
	ret			
_dobit	sbc	a,a		; Set bit=$ff Clear bit=$00
	and	$05		; Set bit=$05 Clear bit=$00
	add	a,$04		; Set bit=$09 Clear bit=$04
	ld	c,a		; Move timer value to C

_OBpulse
	out	($ff),a		; Pulse and delay
	ld	b,$23
	djnz	*
	call	BREAK-1		; Test for SPACE
	ret	nc		; Break

	ld	b,$1e		; Delay
	djnz	*

	dec	c		; Decrement pulse number count
	jr	nz,_OBpulse	; Loop back if more pulses required

	and	a		; Clear carry for next bit test
	djnz	*-1
	jr	_OBbit		; Loop back for next bit





; Enter: HL = Start address of name (terminated by a byte with bit 7 set)
;        DE = Start address of block to save
;	 BC = End address of block to save
;
; Exit: NC = Break pressed

Blk_Load
	ex	de,hl
	ld	(Blk_S),hl
	ld	(Blk_E),bc

_BLnxtp	call	_InByte
	ret	nc
	jr	_BLnxtp




; Get a byte from cassette
_InByte
	ld	c,%00000001	; Prepare 8-bit counter
_IBnxtb	ld	b,$00		; Counter set to 256

_IBtlp	ld	a,$7f		; Read kdb row with
	in	a,($fe)		; the SPACE key
	out	($ff),a		; Output signal to screen
	rra			; Test for key pressed
	ret	nc

	rla			; Reverse last rotation then
	rla			; rotate to get tape test bit
	jr	c,_IBgetbit
	djnz	_IBtlp

	pop	af		; Timeout test
	ld	a,$03		; Force MSB of return address to imitate ROM
	cp	d
	ret	nc

	ld	h,d		; Transfer name start address to HL
	ld	l,e

_InName
	call	_InByte		; Get byte
	bit	7,d		; Is name the null string?
	ld	a,c		; Transfer byte to A
	jr	nz,_IBmatch	; Forward with null string to _IBmatch

	cp	(hl)		; else compare with string in memory
	jr	nz,_BLnxtp	; Go back on mismatch


_IBmatch
	inc	hl		; Address of next character
	rla			; Test for inverted bit
	jr	nc,_InName	; Back if not.

	ld	hl,(Blk_S)
_InProg
	ld	d,b		; Set D to 0 as indicator
	call	_InByte
	ld	(hl),c		; Insert assembled byte
	call	_BlkUpd
	jr	nc,_InProg
	ret


_IBgetbit
	push	de		; Save the timing value
	ld	e,$94
_IBtrail
	ld	b,$1a
_IBcnt	dec	e		; Decrement measuring timer
	in	a,($fe)		; Read tape
	rla
	bit	7,e
	ld	a,e
	jr	c,_IBtrail
	djnz	_IBcnt

	pop	de
	jr	nz,_IBbdone
	cp	$56
	jr	nc,_IBnxtb
_IBbdone
	ccf
	rl	c
	jr	nc,_IBnxtb
	ret



; Test for data block end.
; Enter: HL = address of last output character.
; Exit: NC = more bytes to save
_BlkUpd
	inc	hl
	ex	de,hl
	ld	hl,(Blk_E)
	and	a
	sbc	hl,de
	ex	de,hl
	ret



CURRADD	.word	0
LBcnt	.byte	0
Line25	.word	0



; Scroll screen up one line. Leave cursor at start of line 24

DoScroll .MODULE DoScroll_M
	ld	hl,(D_FILE)
	inc	hl		; Destination
	push	hl
	ld	de,$21
	add	hl,de		; Source
	ld	bc,$2f6
	pop	de
	
	ldir

	ld	hl,(Line25)
	dec	hl
	ld	b,$20
_scrlp	dec	hl
	ld	(hl),0
	djnz	_scrlp
	ld	(CURRADD),hl
	ret





Clear	.MODULE Clear_M
	push	bc
	ld	hl,(D_FILE)
	inc	hl
	ld	(CURRADD),hl
;	ld	(DF_CC),hl

	ld	a,24
_1	ld	b,32
_2	ld	(hl),0
	inc	hl
	djnz	_2
	ld	(hl),HALT
	inc	hl
	dec	a
	jr	nz,_1

	pop	bc
	ret




Line_Del .MODULE Line_Del_M
	ld	a,(IPBcnt)
	and	a
	ret	z
	
	ld	b,a

	ld	hl,(CURRADD)	; Clear cursor
	ld	(hl),0

	push	bc

_1	dec	hl
	call	IsNL
	jr	nz,_2
	dec	hl
_2	ld	(hl),0
	djnz	_1
	ld	(hl),CURSOR
	
	ld	(CURRADD),hl

	pop	bc
	ld	hl,(IPBptr)
_3	dec	hl
	ld	(hl),BLANK
	djnz	_3
	ld	(IPBptr),hl

	xor	a
	ld	(IPBcnt),a
	
	ld	hl,(CURRADD)
	ld	(hl),CURSOR
	ret




; Returns Z if character at HL is a screen newline (HALT)
IsNL
	ld	a,HALT
	cp	(hl)
	ret




DoBS	.MODULE DoBS_M
	ld	a,(IPBcnt)	; Do nothing if buffer empty
	and	a
	ret	z

	ld	hl,(CURRADD)
	ld	(hl),0		; Blank cursor
	dec	hl
	call	IsNL
	jr	nz,_1
	dec	hl
_1	ld	(hl),CURSOR
	ld	(CURRADD),hl

	ld	hl,(IPBptr)
	dec	hl
	ld	(hl),BLANK
	ld	(IPBptr),hl

	ld	hl,IPBcnt
	dec	(hl)
	ret




IPBadd	.word	0		; Input start address
IPBptr	.word	0		; Pointer within input area
IPBmax	.byte	0		; Maximum number of bytes to read
IPBcnt	.byte	0		; Bytes read so far

	.MODULE LBInput_M
LBInput	ld	hl,LBuff	; Set up input buffer addresses for Line
	ld	(IPBadd),hl	; Buffer use
	ld	(IPBptr),hl
	ld	a,LBSIZE
	ld	(IPBmax),a

	ld	b,a		; Clear the Line Buffer
_1	ld	(hl),BLANK
	inc	hl
	djnz	_1

	call	CRLF
	call	Input

	ld	hl,(IPBptr)	; Reload the line buffer pointer
	ld	(LBP),hl

	ret




; EXPECT input entry
	.MODULE EXPECTI_M
EXInput
	call	PopOne
	xor	a
	ld	(IPBcnt),a
	bit	7,h
	jr	z,_2
_1	call	PopOne
	ret
_2	cp	l
	jr	z,_1

	ld	a,l
	ld	(IPBmax),a
	call	PopOne
	ld	(IPBadd),hl
	ld	(IPBptr),hl

	call	Input

	ret




Input	.MODULE Input_M
	ld	hl,(CURRADD)
	ld	(hl),CURSOR
	ld	hl,IPBcnt
	ld	(hl),0

_dosc	call	ScanRepeat	; HL preserved

	cp	NL
	jr	nz,_tryRUB
	ld	(hl),0
	call	CRLF
	ld	hl,(IPBadd)
	ld	(IPBptr),hl
;	ld	hl,IPBcnt
;	ld	(hl),0
	ret

_tryRUB	cp	BS
	jr	nz,_try_ldel
	call	DoBS
	jr	_dosc

_try_ldel
	cp	DELLINE
	jr	nz,_newchr
	call	Line_Del
	jr	_dosc


_newchr	push	af
	ld	a,(IPBcnt)	; Buffer full?
	ld	hl,IPBmax
	cp	(hl)
	jr	nz,_notf

	ld	hl,(CURRADD)
	dec	hl
	call	IsNL
	jr	nz,_2
	dec	hl
_2	pop	af
	push	af
	call	Asc2zx
	ld	(hl),a		; Screen
	pop	af

	ld	hl,(IPBptr)	; Overwrite last buff char
	dec	hl
	ld	(hl),a		; Not screen
	jr	_dosc

_notf
	ld	hl,IPBcnt
	inc	(hl)

	pop	af
	ld	hl,(IPBptr)	; Store in Buff
	ld	(hl),a		; Not screen
	inc	hl
	ld	(IPBptr),hl

	call	PrtChr
	ld	a,CURSOR
	call	PrtChr
	ld	hl,(CURRADD)
	dec	hl
	call	IsNL
	jr	nz,_3
	dec	hl
_3	ld	(hl),CURSOR
	ld	(CURRADD),hl

	
	jr	_dosc







; Returns Z if print position at Line 25
ScrollChk
	.MODULE ScrollChk_M
	ld	hl,(CURRADD)
	push	hl
	ld	de,(Line25)
	and	a
	sbc	hl,de
	pop	hl
	ret
	

; Perform carriage return, scrolling if necessary
CRLF	.MODULE CR_M
	push	hl
	push	de
	push	bc

	call	ScrollChk
	jr	nz,_1
	call	DoScroll
	jr	_4

_1	ld	hl,(CURRADD)
	ld	a,(hl)
	cp	CURSOR
	jr	nz,_2
	ld	(hl),0

_2	ld	a,(hl)
	cp	HALT
	jr	z,_3
	inc	hl
	jr	_2

_3	inc	hl
	ld	(CURRADD),hl
	call	ScrollChk
	jr	nz,_4
	call	DoScroll
_4	pop	bc
	pop	de
	pop	hl
	ret

	


; Print character at address (CURRADD), scrolling if necessary 
; Increment character position
PrtChr	.MODULE PrtChr_M
	push	hl
	push	de
	push	bc
	call	ScrollChk
	jr	nz,_1
	call	DoScroll

_1	call	Asc2zx
	ld	(hl),a

	inc	hl
	call	IsNL
	jr	nz,_2
	inc	hl
_2	ld	(CURRADD),hl
	pop	bc
	pop	de
	pop	hl
	ret



; Input: Buffer number in A
; Output: Returns Z if buffer found in mass storage. If so HL contains
; the address of the first byte of the buffer header.

Findbuff .MODULE Findbuff_M
	ld	hl,Buff1
	cp	(hl)
	ret	z
	ld	hl,Buff2
	cp	(hl)
	ret




; Input: Address of current buffer in HL
; Output: Address of next buffer in HL
	.MODULE Nextbuff_M
Nextbuff
	ld	de,Buff1
	and	a
	sbc	hl,de
	ld	hl,Buff2
	jr	z,_1
	ld	hl,Buff1
_1	ret




; Print value in A as a two digit decimal number
	.MODULE DecPrt_M
DecPrt
	ld	c,-10
	call	_1
	ld	c,-1
_1	ld	b,'0'-1
_2	inc	b
	add	a,c
	jr	c,_2
	sub	c
	push	af
	ld	a,b
	call	PrtChr
	pop	af
	ret




; Convert binary in a to two printable hex numbers in d & e.
Bin2Hex	.MODULE Bin2Hex_M
	push	af
	rra
	rra
	rra
	rra
	call	_1
	ld	d,a
	pop	af
_1	and	$0f
	cp	$0a
	jr	c,_2
	add	a,$07
_2	add	a,$30
	ld	e,a
	ret




; Set up tape block filename. Enter with screen number in A
	
MkFname	
	call	Bin2Hex

	ld	hl,Fname+3
	ld	a,d
	call	Asc2zx
	ld	(hl),a
	inc	hl
	ld	a,e
	call	Asc2zx
	ld	(hl),a

	ld	hl,AFname+4
	ld	(hl),d
	inc	hl
	ld	(hl),e
	
	ret


; Entry point
; Do preliminary setup bofoe entering FORTH via Do_Forth

Start
	call	Clear

	ld	hl,$0000	; Save original return stack value
	ld	(Scr),hl	; Reset SCR, WARNING and R#
	ld	(Warn),hl
	ld	(Rhash),hl
	add	hl,sp		; Save current stack pointer
	ld	(StkIni),hl

	ld	hl,(D_FILE)	; Calculate start of line 25
	ld	de,$319		; (used to test if scrolling required)
	add	hl,de
	ld	(Line25),hl


	ld	hl,LBuff	; Reset line buffer start and pointer
	ld	(LBP),hl
	ld	(LBPbase),hl
	ld	hl,(D_FILE)	; Reset print address
	inc	hl
	ld	(CURRADD),hl

	xor	a		; Reset count of characters in line buffer
	ld	(LBcnt),a


	ld	hl,DStack	; Reset the computation stack
	ld	(DSP),hl


#IFDEF UDG
	call	Set_CHR128	; Setup UDGs
#ENDIF

	xor	a		; Resetting the numeric base is a marker
	ld	(Base),a	; for a cold start
	jp	Do_Forth	; Start Forth.




; Convert an ASCII character code to the equivalent ZX81 screen code. 
Asc2zx	.MODULE Asc2zx_M
	cp	$20
	jr	c,_1
	cp	$a0
	jr	c,_2
_1	xor	a
	ret
_2	push	hl
	sub	BLANK

	ld	hl,AsciiConv
	add	a,l
	ld	l,a
	jr	nc,_3
	inc	h
_3
	ld	a,(hl)
	pop	hl
	ret




Keytables
#IFDEF UDG
; 007E-00A4 (unshifted)
	.byte	"ZXCVASDFGQWERT1234509876POIUY",NL,"LKJH .MNB"
						  ; $22=double quote
; 00A5-00CB (shifted).
	.byte	':',$3b,"?/"		          ; ZXCV $3b=semicolon
	.byte	'@',0,0,0,0			  ; ASDFG
	.byte	"&^!%'"				  ; QWERT
	.byte	DELLINE,"[]#",CURL		  ; 12345
	.byte	BS,GRAPH,CURR,CURU,CURD		  ; 09876
	.byte	$22,")($",0			  ; POIUY  $22="
	.byte	0,"=+-#"			  ; Newline LKJH
	.byte	$5f,",><*"			  ; Space .MNB $5f=
#ELSE
; 007E-00A4 (unshifted)
	.byte	"ZXCVASDFGQWERT1234509876POIUY",NL,"LKJH .MNB"
						  ; $22=double quote
; 00A5-00CB (shifted).
	.byte	':',$3b,"?/"		          ; ZXCV $3b=semicolon
	.byte	'0',0,0,0,0			  ; ASDFG
	.byte	0,0,0,0,0				  ; QWERT
	.byte	DELLINE,0,0,0,CURL		  ; 12345
	.byte	BS,GRAPH,CURR,CURU,CURD		  ; 09876
	.byte	$22,")($",0			  ; POIUY  $22="
	.byte	0,"=+-#"			  ; Newline LKJH
	.byte	$5f,",><*"			  ; Space .MNB $5f=
#ENDIF



; Mapping of extra punctuation characters
; @=A [=2 ]=3 #=4 (also H) &=Q ^=W '=T !=E %=R
; @='A't  !='E'xclaim '='T'ick #='H'ash (also looks a bit like a '4')




#IFDEF UDG
; Set up CHR$128 mode with chardefs for #![]@&%^'
Set_CHR128
	ld	hl,$1e00
	ld	de,$2000
	ld	bc,$200
	ldir

	ld	hl,CharDefs
	ld	bc,CdEnd-CharDefs
	ld	de,$2008
	ldir

	ld	a,$20
	ld	i,a

	ret



CharDefs
	.byte	$00,$3c,$42,$5e,$4e,$40,$3c,$00	; @
	.byte	$00,$08,$08,$08,$08,$00,$08,$00	; !
	.byte	$00,$08,$08,$00,$00,$00,$00,$00 ; '
	.byte	$00,$24,$24,$ff,$24,$ff,$24,$24 ; #
	.byte	$00,$1c,$10,$10,$10,$10,$1c,$00 ; [
	.byte	$00,$38,$08,$08,$08,$08,$38,$00 ; ]
	.byte	$00,$10,$28,$10,$2a,$44,$3a,$00 ; &
	.byte	$00,$62,$64,$08,$10,$26,$46,$00 ; %
	.byte	$00,$00,$40,$20,$10,$08,$04,$00 ; \
	.byte	$00,$10,$28,$44,$00,$00,$00,$00 ; ^
CdEnd
#ENDIF




AsciiConv
#IFDEF UDG	
	.byte	$00,$02,$0b,$04,$0d,$08,$07,$03,$10,$11,$17,$15,$1a,$16,$1b,$18
	.byte	$1c,$1d,$1e,$1f,$20,$21,$22,$23,$24,$25,$0e,$19,$13,$14,$12,$0f
	.byte	$01,$26,$27,$28,$29,$2a,$2b,$2c,$2d,$2e,$2f,$30,$31,$32,$33,$34
	.byte	$35,$36,$37,$38,$39,$3a,$3b,$3c,$3d,$3e,$3f,$05,$09,$06,$0a,$0c

	; $81 + inverse A-O
	.byte	$81,$a6,$a7,$a8,$a9,$aa,$ab,$ac,$ad,$ae,$af,$b0,$b1,$b2,$b3,$b4
	; Inverse P-Z [\]^
	.byte	$b5,$b6,$b7,$b8,$b9,$ba,$bb,$bc,$bd,$be,$bf,$85,$89,$86,$8a,$8c
	; Inverse  !"# cr %&'()*+,-./
	.byte	$80,$82,$8b,$84,$8d,$88,$87,$83,$90,$91,$97,$95,$9a,$96,$9b,$98
	; Inverse 0-9 :;<=>?
	.byte	$9c,$9d,$9e,$9f,$a0,$a1,$a2,$a3,$a4,$a5,$8e,$99,$93,$94,$92,$8f
#ELSE
	.byte	$00,$00,$0b,$00,$0d,$00,$00,$00,$10,$11,$17,$15,$1a,$16,$1b,$18
	.byte	$1c,$1d,$1e,$1f,$20,$21,$22,$23,$24,$25,$0e,$19,$13,$14,$12,$0f
	.byte	$01,$26,$27,$28,$29,$2a,$2b,$2c,$2d,$2e,$2f,$30,$31,$32,$33,$34
	.byte	$35,$36,$37,$38,$39,$3a,$3b,$3c,$3d,$3e,$3f,$00,$00,$00,$00,$0c

	; $81 + inverse A-O
	.byte	$81,$a6,$a7,$a8,$a9,$aa,$ab,$ac,$ad,$ae,$af,$b0,$b1,$b2,$b3,$b4
	; Inverse P-Z [\]^
	.byte	$b5,$b6,$b7,$b8,$b9,$ba,$bb,$bc,$bd,$be,$bf,$80,$80,$80,$80,$8c
	; Inverse  !"# cr %&'()*+,-./
	.byte	$80,$80,$8b,$80,$8d,$80,$80,$80,$90,$91,$97,$95,$9a,$96,$9b,$98
	; Inverse 0-9 :;<=>?
	.byte	$9c,$9d,$9e,$9f,$a0,$a1,$a2,$a3,$a4,$a5,$8e,$99,$93,$94,$92,$8f
#ENDIF


last
	
DicTop

	.END
