/* SPLIT.C - a program to split TASM obj files produced using multiple
** .org statements into individual files. Output filenames have the format
** "hexaddr.bin". Input files to the program are the OBJ file and the LST
** file. The LST file is parsed to calculate the sections of the OBJ file
** that need extracting.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAXLINE 512


void save_section(int,int,int,FILE*);




int main(int argc, char **argv)
{
    char listfn[80];
    char objfn[80];
    char line[MAXLINE];

    FILE *listf;
    FILE *objf;
    
    int  ostart;
    int  oend;
    
    int  odisp;
    int  orgaddr;

    int last;
    

    char directive[MAXLINE];
    char *res;
    int  prevorg;
    


    if(argc < 2)
    {
        fprintf(stderr,"usage: split basename\n");
        exit(-1);
    }

    strcpy(listfn,argv[1]);
    strcat(listfn,".lst");

    strcpy(objfn,argv[1]);
    strcat(objfn,".obj");

    listf = fopen(listfn,"rb");
    if(!listf)
    {
        fprintf(stderr,"Cannot open list file %s\n",listfn);
        exit(-1);
    }

    objf = fopen(objfn,"rb");
    if(!objf)
    {
        fprintf(stderr,"Cannot open object file %s\n",objfn);
        exit(-1);
    }


    /* Get to first .org statement */
    while((res = fgets(line,MAXLINE,listf)) != NULL)
    {
        if(sscanf(line,"%*x%x%s",&orgaddr,directive) != 2)
            continue;

        if(!strcmp(directive,".org"))
            break;
    }

    if(!res)
    {
        fprintf(stderr,"No .org statement\n");
        exit(-1);

    }

    prevorg = orgaddr;
    
    odisp = orgaddr;
    ostart = 0;
    

    while((res = fgets(line,MAXLINE,listf)) != NULL)
    {
        if(sscanf(line,"%*x%x%s",&orgaddr,directive) != 2)
            continue;

        if(!strcmp(directive,".org") || !strcmp(directive,".end"))
        {
            oend = last - odisp;
            save_section(ostart,oend,prevorg,objf);
            ostart = orgaddr - odisp;
            prevorg = orgaddr;
        }
        else
            last = orgaddr;
        
        if(!strcmp(directive,".end"))
            break;
        
        
    }


    fclose(listf);
    fclose(objf);

    return 0;
}




void save_section(int ostart,int oend,int orgaddr,FILE *objf)
{
    int len;
    char fname[MAXLINE];
    char buf[65536];
    FILE *outf;
    
    len = (oend - ostart) + 1;

    sprintf(fname,"%x.bin",orgaddr);
    outf = fopen(fname,"wb");
    if(!outf)
    {
        fprintf(stderr,"Cannot open output file %s\n",fname);
        exit(-1);
    }


    fseek(objf,(size_t)ostart,SEEK_SET);
    fread((void*)buf,(size_t)1,(size_t)len,objf);
    fwrite((void*)buf,(size_t)1,(size_t)len,outf);
    
    printf("%x.bin %x %x Length=%x\n",orgaddr,ostart,oend,len);
    fclose(outf);
    

    return;
}
